
/********************

PhyloBayes MPI. Copyright 2010-2013 Nicolas Lartillot, Nicolas Rodrigue, Daniel Stubbs, Jacques Richer.

PhyloBayes is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License
as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
PhyloBayes is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

See the GNU General Public License for more details. You should have received a copy of the GNU General Public License
along with PhyloBayes. If not, see <http://www.gnu.org/licenses/>.

**********************/

#ifndef BIOLOGICALSEQUENCES_H
#define BIOLOGICALSEQUENCES_H

#include <cstdlib>
#include <iostream>
#include <string>
using namespace std;


	static const string Path = "";

	enum DataType {DNA=0,RNA=1,Protein=2,Other=3};

	const char AAset[] = {'A','C','D','E','F','G','H','I','K','L','M','N','P','Q','R','S','T','V','W','Y','a','c','d','e','f','g','h','i','k','l','m','n','p','q','r','s','t','v','w','y','-','?','$','.','B','Z','*','X','x'};
	const int AAN = 49;
	const int DNAN = 37;
	const int RNAN = 37;
	const char DNAset[] = {'A','C','G','T','a','c','g','t','B','D','H','K','M','N','R','S','V','W','Y','b','d','h','k','m','n','r','s','v','w','y','-','?','$','.','*','X','x'};
	const char RNAset[] = {'A','C','G','U','a','c','g','u','B','D','H','K','M','N','R','S','V','W','Y','b','d','h','k','m','n','r','s','v','w','y','-','?','$','.','*','X','x'};


// amino acids

	const int precision = 10000;
	const string Alphabet = "Amino_Acids";
	const char AminoAcids[] = {'A','C','D','E','F','G','H','I','K','L','M','N','P','Q','R','S','T','V','W','Y','-'};
	const char aminoacids[] = {'a','c','d','e','f','g','h','i','k','l','m','n','p','q','r','s','t','v','w','y','-'};
	const char DNAletters[] = {'A','C','G','T'};
	const char dnaletters[] = {'a','c','g','t'};
	const char RNAletters[] = {'A','C','G','U'};
	const char rnaletters[] = {'a','c','g','u'};
	
	const int Dayhoff6Table[] = {3,5,2,2,4,3,1,0,1,0,0,2,3,2,1,3,3,0,4,4};
	const int Dayhoff4Table[] = {3,-1,2,2,0,3,1,0,1,0,0,2,3,2,1,3,3,0,2,2};
	const int unknown = -1;


	const int Naa = 20;
	const int Nnuc = 4;
	const int Ncodon = 64;
	const string Codons[] = {"TTT","TTC","TTA","TTG","TCT","TCC","TCA","TCG","TAT","TAC","TAA","TAG","TGT","TGC","TGA","TGG","CTT","CTC","CTA","CTG","CCT","CCC","CCA","CCG","CAT","CAC","CAA","CAG","CGT","CGC","CGA","CGG","ATT","ATC","ATA","ATG","ACT","ACC","ACA","ACG","AAT","AAC","AAA","AAG","AGT","AGC","AGA","AGG","GTT","GTC","GTA","GTG","GCT","GCC","GCA","GCG","GAT","GAC","GAA","GAG","GGT","GGC","GGA","GGG"};
	const int codonpos[][64] = {
	{3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2},
	{3,3,3,3,1,1,1,1,0,0,0,0,2,2,2,2,3,3,3,3,1,1,1,1,0,0,0,0,2,2,2,2,3,3,3,3,1,1,1,1,0,0,0,0,2,2,2,2,3,3,3,3,1,1,1,1,0,0,0,0,2,2,2,2},
	{3,1,0,2,3,1,0,2,3,1,0,2,3,1,0,2,3,1,0,2,3,1,0,2,3,1,0,2,3,1,0,2,3,1,0,2,3,1,0,2,3,1,0,2,3,1,0,2,3,1,0,2,3,1,0,2,3,1,0,2,3,1,0,2}};


	enum GeneticCodeType {Universal = 0, MtMam = 1, MtInv = 2};
	// universal genetic code
	// const string UniStopCodons[] = {"TAA","TAG","TGA"};
	const int UniNStopCodons = 3;
	const int UniStopCodons[] = {10,11,14};
	int const UniCodonCode[] = {4,4,9,9,15,15,15,15,19,19,-1,-1,1,1,-1,18,9,9,9,9,12,12,12,12,6,6,13,13,14,14,14,14,7,7,7,10,16,16,16,16,11,11,8,8,15,15,14,14,17,17,17,17,0,0,0,0,2,2,3,3,5,5,5,5};
	const int UniStopPos1[] = {3,3,3};
	const int UniStopPos2[] = {0,0,2};
	const int UniStopPos3[] = {0,2,0};

	const int MtInvNStopCodons = 2;
	const int MtInvStopCodons[] = {10,11};
	int const MtInvCodonCode[] = {4,4,9,9,15,15,15,15,19,19,-1,-1,1,1,18,18,9,9,9,9,12,12,12,12,6,6,13,13,14,14,14,14,7,7,10,10,16,16,16,16,11,11,8,8,15,15,15,15,17,17,17,17,0,0,0,0,2,2,3,3,5,5,5,5};
	const int MtInvStopPos1[] = {3,3};
	const int MtInvStopPos2[] = {0,0};
	const int MtInvStopPos3[] = {0,2};

	// mammal mitochondrial genetic code
	const int MtMamNStopCodons = 4;
	const int MtMamStopCodons[] = {10,11,46,47};
	int const MtMamCodonCode[] = {4,4,9,9,15,15,15,15,19,19,-1,-1,1,1,18,18,9,9,9,9,12,12,12,12,6,6,13,13,14,14,14,14,7,7,10,10,16,16,16,16,11,11,8,8,15,15,-1,-1,17,17,17,17,0,0,0,0,2,2,3,3,5,5,5,5};
	const int MtMamStopPos1[] = {3,3,0,0};
	const int MtMamStopPos2[] = {0,0,2,2};
	const int MtMamStopPos3[] = {0,2,0,2};

	inline istream& operator>>(istream& is, GeneticCodeType& type)	{
		string t;
		is >> t;
		if (t == "Universal")	{
			type = Universal;
		}
		else if (t == "MtMam")	{
			type = MtMam;
		}
		else if (t == "MtInv")	{
			type = MtInv;
		}
		else	{
			cerr << "error in istream genetic code type\n";
			cerr << t << '\n';
			exit(1);
		}
		return is;
	}

	inline ostream& operator<<(ostream& os, GeneticCodeType type)	{
		if (type == Universal)	{
			os << "Universal\n";
		}
		else if (type == MtMam)	{
			os << "MtMam\n";
		}
		else if (type == MtInv)	{
			os << "MtInv\n";
		}
		else	{
			cerr << "error in ostream genetic code type\n";
			cerr << (int) type << '\n';
			exit(1);
		}
		return os;
	}


	const double WAG_RR[] = { 1.02704,0.738998,1.58285,0.210494,1.41672,0.316954,0.193335,0.906265,0.397915,0.893496,0.509848,1.43855,0.908598,0.551571,3.37079,2.12111,2.00601,0.113133,0.240735,0.0302949,0.021352,0.39802,0.306674,0.248972,0.170135,0.0740339,0.384287,0.390482,0.265256,0.109404,0.0988179,0.528191,1.40766,0.512984,1.00214,0.71707,0.543833,6.17416,0.0467304,0.865584,0.930676,0.039437,0.479855,0.0848047,0.103754,5.42942,0.423984,0.616783,0.147304,1.07176,0.374866,0.152335,0.129767,0.325711,0.0811339,0.567717,0.570025,0.127395,2.58443,0.154263,0.315124,0.947198,0.682355,5.46947,0.439157,0.704939,0.822765,0.588731,0.156557,0.196303,0.049931,0.679371,1.05947,0.088836,2.11517,1.19063,0.0961621,0.161444,0.0999208,0.102711,0.545931,0.171903,0.649892,1.52964,6.45428,0.24941,0.0304501,0.373558,0.0613037,0.1741,1.12556,0.24357,0.330052,0.584665,1.34182,0.225833,0.187247,0.336983,0.103604,0.13819,0.890432,0.499462,0.404141,3.95629,0.696198,4.29411,2.13715,0.740169,0.473307,0.118358,0.262569,3.87344,0.323832,3.17097,4.25746,0.554236,0.0999288,0.113917,0.186979,0.31944,1.45816,7.8213,0.212483,0.42017,0.257555,0.934276,3.01201,0.556896,3.8949,5.35142,0.96713,1.38698,0.305434,0.137505,0.133264,4.85402,0.131528,0.415844,0.869489,0.497671,0.344739,0.326622,1.80034,0.665309,0.398618,0.198221,0.171329,1.54526,0.683162,0.493905,1.51612,2.05845,0.515706,0.428437,0.195081,1.54364,0.635346,3.97423,2.03006,0.196246,0.0719167,1.086,0.933372,0.679489,1.61328,0.795384,0.314887,0.139405,0.216046,3.0355,1.02887,0.857928,0.301281,0.215737,0.22771,1.22419,0.554413,0.251849,1.16392,0.381533,4.37802,0.232739,0.523742,0.786993,1.38823,0.110864,0.291148,0.365369,0.31473,2.48539 };

	const double JTT_RR[] = { 56,81,105,15,179,27,36,35,30,54,54,194,57,58,378,475,298,9,11,10,5,78,59,69,17,7,23,31,34,14,9,113,223,42,62,115,209,767,4,130,112,11,26,7,15,528,15,49,16,59,38,31,4,46,5,119,26,12,181,9,18,58,18,323,29,30,32,45,10,7,5,40,89,4,248,43,10,17,4,5,92,12,62,53,536,23,6,27,6,14,81,24,26,137,201,33,47,55,8,16,45,56,33,391,115,597,328,73,46,11,8,573,21,229,479,47,10,9,22,40,245,961,9,32,14,65,263,21,292,646,47,103,14,10,8,388,12,102,72,38,59,25,180,52,24,30,16,43,44,29,226,323,24,18,15,86,45,503,232,16,8,70,164,74,285,118,23,6,10,310,53,51,20,18,24,101,64,17,126,20,477,38,35,63,112,12,21,25,16,71};

	const double JTT_Stat[] = {
0.076748
,0.019803
,0.051544
,0.061830
,0.040126
,0.073152
,0.022944
,0.053761
,0.058676
,0.091904
,0.023826
,0.042645
,0.050901
,0.040752
,0.051691
,0.068765
,0.058565
,0.066005
,0.014261
,0.032102
};

const double WAG_Stat[] = {
0.0866279
,0.0193078
, 0.0570451
, 0.0580589 
, 0.0384319 
, 0.0832518 
, 0.0244313
, 0.048466  
, 0.0620286 
, 0.086209
, 0.0195027 
, 0.0390894 
, 0.0457631 
, 0.0367281 
, 0.043972
, 0.0695179
, 0.0610127
, 0.0708956
, 0.0143859
, 0.0352742
};

const double mtREV_Stat[] = {
0.0692,
0.0065,
0.0186,
0.0236,
0.0611,
0.0557,
0.0277,
0.0905,
0.0221,
0.1675,
0.0561,
0.04,
0.0536,
0.0238,
0.0184,
0.0725,
0.087,
0.0428,
0.0293,
0.034};


	const double mtREV_RR[] = { 59.93,17.67,9.77,6.37,120.71,13.9,96.49,8.36,25.46,141.88,26.95,54.31,1.9,23.18,387.86,480.72,195.06,1.9,6.48,1.9,1.9,70.8,30.71,141.49,62.73,1.9,25.65,6.18,58.94,31.26,75.24,103.33,277.05,179.97,1.9,33.6,254.77,583.55,4.98,56.77,113.99,4.34,2.31,1.9,1.9,794.38,13.43,55.28,1.9,69.02,28.01,1.9,19.86,21.21,2.67,28.28,49.12,3.31,313.86,1.9,1.9,63.05,12.83,313.56,1.9,54.71,14.82,21.14,1.9,13.12,1.9,48.16,84.67,6.44,216.06,90.82,15.2,17.31,19.11,4.69,64.29,33.85,6.35,7.84,465.58,1.9,5.98,22.73,2.41,1.9,53.3,1.9,6.75,23.03,125.93,11.17,2.53,10.92,3.21,12.26,127.67,11.49,11.97,496.13,60.97,582.4,165.23,77.46,44.78,1.9,7.08,670.14,19.57,329.09,517.98,27.1,20.63,8.34,1.9,47.7,368.43,1222.94,1.9,25.01,14.88,91.37,608.7,50.1,465.58,141.4,105.79,136.33,1.9,24,51.17,537.53,15.16,40.1,39.7,15.58,73.61,126.4,91.67,32.44,44.15,65.41,18.84,47.37,1.9,111.16,528.17,387.54,21.71,39.96,73.31,173.56,13.24,494.39,238.46,1.9,10.68,191.36,137.29,23.64,169.9,128.22,8.23,4.21,16.21,220.99,54.11,94.93,19,1.9,38.82,6.04,2.08,7.64,21.95,1.9,597.21,1.9,38.58,64.92,204.54,9.99,38.73,5.37,1.9,26.25};

const double mtZOA_RR[] = { 272.5,16.1,17.1,26.5,289.3,2.3,33.2,0.2,15.6,136.5,1.7,61.8,7.3,3.3,644.9,378.1,321.9,3.1,2.1,9.5,3.4,167.8,82.4,37.5,48.8,0.2,59.7,264.8,94.6,8.1,49.1,61.1,656.3,199,243.8,36.1,76.7,883.6,2,26,55.6,1.5,0.2,0.5,0.2,617,0.2,19.3,3.2,51.4,2.3,3.7,0.2,13.9,0.2,43.1,14.9,7.3,163.2,1,0.2,174,20.6,349.4,6.4,38.4,34.5,16.3,3.5,10,5.8,13.1,90.3,16.3,234.2,215.6,12.9,10.5,9.5,0.2,32.2,19.3,59.6,58.2,590.5,7.4,3.4,10.1,3.5,52.2,99.3,1.3,8.9,7.2,257.1,5.2,23.7,12.3,4.3,1.6,23.9,6.6,7.1,228.9,15.6,421.8,61.7,23.1,19.4,0.3,4.5,266.5,8.4,425.2,449.7,24.3,2.6,0.2,0.2,7.2,222.3,1710.6,9.7,13.1,6.7,83,413.4,24.3,334,292.3,144.9,75.5,11.1,6.6,45,636.3,7.9,11.4,23,4.1,15.2,50,126.1,27.2,5.7,73.7,5.4,83.9,3.8,95.3,305.1,279.6,48.7,41.4,22.6,190.3,33.6,420.2,184.6,7.1,6.4,141.6,52.2,7.5,79.7,56.9,17.9,1.3,4.2,231,96.4,39.4,9,6.1,52.3,11.8,3.2,5.1,16.9,13.8,666.3,49.5,10.3,29.7,396.4,3.6,29,13.7,15.6,79.8};

const double mtART_RR[] = { 254.0,1.0,0.2,13.0,200.0,0.2,26.0,0.2,4.0,121.0,0.2,49.0,0.2,0.2,673.0,244.0,340.0,0.2,1.0,11.0,0.2,184.0,81.0,12.0,63.0,0.2,79.0,312.0,98.0,0.2,0.2,36.0,664.0,183.0,350.0,22.0,72.0,862.0,0.2,12.0,0.2,7.0,2.0,1.0,0.2,500.0,0.2,0.2,4.0,44.0,0.2,0.2,0.2,0.2,0.2,44.0,15.0,7.0,106.0,2.0,0.2,183.0,8.0,262.0,0.2,31.0,43.0,14.0,11.0,8.0,1.0,14.0,118.0,11.0,263.0,322.0,20.0,15.0,0.2,5.0,36.0,14.0,52.0,54.0,792.0,0.2,3.0,0.2,1.0,56.0,121.0,0.2,3.0,0.2,226.0,0.2,3.0,2.0,9.0,0.2,0.2,6.0,0.2,180.0,1.0,314.0,41.0,11.0,19.0,0.2,0.2,191.0,3.0,515.0,515.0,21.0,0.2,11.0,2.0,7.0,204.0,1855.0,0.2,12.0,4.0,106.0,467.0,17.0,349.0,209.0,144.0,70.0,26.0,16.0,117.0,885.0,13.0,12.0,16.0,2.0,8.0,48.0,85.0,21.0,20.0,79.0,5.0,67.0,5.0,112.0,289.0,281.0,71.0,71.0,17.0,262.0,0.2,398.0,166.0,23.0,8.0,251.0,39.0,0.2,87.0,47.0,32.0,0.2,18.0,154.0,52.0,44.0,0.2,7.0,87.0,3.0,0.2,0.2,0.2,4.0,660.0,61.0,2.0,30.0,544.0,0.2,46.0,0.2,2.0,38.0};

const double mtZOA_Stat[] = {
0.06888
,0.009966
,0.020696
,0.024989
,0.081712
,0.071968
,0.026814
,0.085072
,0.019276
,0.156717
,0.050652
,0.03039
,0.044803
,0.018623
,0.021037
,0.080535
,0.056386
,0.066083
,0.027998
,0.037404
};

const double mtART_Stat[] = {
0.054116
,0.009709
,0.02016
,0.024289
,0.088668
,0.068183
,0.024518
,0.092639
,0.021718
,0.148658
,0.061453
,0.039903
,0.041826
,0.018781
,0.018227
,0.09103
,0.049194
,0.057701
,0.029786
,0.039443
};



const double WAGHSSP_Stat[] = {
0.085243,
0.013379,
0.056805,
0.061844,
0.043623,
0.089244,
0.023781,
0.056174,
0.048492,
0.09882,
0.020018,
0.037861,
0.052087,
0.035058,
0.051938,
0.050841,
0.049985,
0.06879,
0.019125,
0.036891};

const double WAGHSSP_RR[] = {1.64948,0.739217,1.79304,0.142332,1.24831,0.396223,0.159376,1.05487,0.280498,0.718381,0.500177,1.31155,1.35677,0.592968,4.43919,2.25227,1.90408,0.0726856,0.162823,0.0512417,0.0242946,0.533782,0.308989,0.483028,0.226929,0.0929962,0.438247,0.613013,0.456734,0.0906867,0.196397,0.726524,2.55884,1.02353,1.41307,0.323252,0.778724,6.34429,0.0230597,0.72266,0.899546,0.0210508,0.562634,0.0356064,0.0652733,4.89097,0.409282,0.859458,0.167388,1.20123,0.477169,0.0746718,0.0454355,0.148609,0.0386372,0.424668,0.573973,0.0967001,2.88655,0.118742,0.244628,0.937013,0.651387,5.80882,0.559445,0.931577,0.985959,0.436752,0.0677616,0.134337,0.0237905,0.706362,0.723098,0.0672424,1.63593,1.24169,0.0812053,0.0565816,0.0917577,0.065748,0.320404,0.12655,0.431432,1.28087,6.50665,0.200339,0.0100586,0.331754,0.0241816,0.0735595,1.03696,0.149487,0.310133,0.319571,1.21641,0.15982,0.0685892,0.0745881,0.0359701,0.13623,1.10918,0.390761,0.343367,4.00467,0.405584,4.76286,2.26409,0.864915,0.538271,0.155706,0.351174,4.1817,0.293201,3.13648,3.95457,0.218601,0.0680664,0.159065,0.138131,0.159465,1.15942,8.91365,0.0723499,0.239056,0.279794,0.965896,3.03521,0.535245,5.59335,7.09493,1.21047,1.78078,0.33551,0.0728535,0.155629,5.42731,0.102376,0.218293,0.857077,0.367929,0.219897,0.313495,1.65183,0.335283,0.334029,0.298733,0.10891,1.57295,0.456507,0.414598,1.61871,1.85172,0.371374,0.38636,0.152446,2.00206,0.675117,4.04362,2.0282,0.134363,0.037322,0.577698,0.757535,0.35979,1.34685,0.597446,0.249979,0.044876,0.074377,3.47976,1.46085,1.30967,0.379146,0.168593,0.260319,1.01135,0.684154,0.214619,0.353412,0.242479,6.10033,0.215848,0.209496,0.384242,1.76289,0.0769942,0.197169,0.126458,0.217706,1.7447};

const double LG_Stat[] = {
0.079066,
0.0129369,
0.0530516,
0.0715863,
0.0423017,
0.0573372,
0.0223546,
0.0621565,
0.0646003,
0.099081,
0.0229506,
0.0419774,
0.0440395,
0.0407668,
0.0559413,
0.0611971,
0.0532871,
0.0691469,
0.0120656,
0.0341554
};

const double LG_RR[] = {2.43501,0.386559,1.01598,0.248188,2.02114,0.35106,0.146574,0.524859,0.386747,1.09961,0.270803,1.15206,0.94882,0.415855,4.62446,2.09301,2.49251,0.176789,0.214201,0.0611966,0.00342322,1.08123,0.556892,0.626623,0.313658,0.0129776,0.581098,0.874258,0.517278,0.0737435,0.0829654,0.522935,2.72396,1.11863,1.91672,0.655562,1.1402,5.12993,0.0170374,0.826565,0.906972,0.010458,0.27681,0.014748,0.0249932,4.96585,0.385885,0.512014,0.121261,1.21333,0.416606,0.0371423,0.0292405,0.132172,0.0184023,0.341267,0.414673,0.0433031,1.7679,0.0681587,0.16996,0.529941,0.410296,4.03888,0.356061,0.598676,0.591407,0.23971,0.0761602,0.117429,0.0876389,0.667316,1.08855,0.0233983,2.53635,1.75976,0.0875798,0.0924114,0.035076,0.0515759,0.353957,0.161416,0.640457,2.40372,7.63435,0.304716,0.00851629,0.290191,0.0432987,0.136505,1.40641,0.192681,0.262136,0.381714,1.70218,0.127015,0.0750342,0.262657,0.0534908,0.106516,0.682111,0.358356,0.432853,4.41125,0.497794,4.70891,2.37387,0.968501,0.571566,0.116427,0.584078,5.19151,0.155611,4.055,4.18072,0.187342,0.0765801,0.0712711,0.124231,0.0627125,1.01127,10.4177,0.109234,0.22747,0.134512,0.642335,2.09847,0.381839,3.16401,6.1886,0.732413,1.11216,0.181177,0.048821,0.129065,6.17517,0.0669398,0.243648,0.5698,0.295289,0.178326,0.296353,1.66574,0.606165,0.293136,0.362942,0.0976794,1.63622,0.473613,0.339422,1.97647,1.85746,0.681046,0.470848,0.158271,1.6589,0.73554,3.92125,1.95721,0.081869,0.0443899,0.598725,0.610728,0.325308,1.30906,0.559051,0.290058,0.0930628,0.0876663,2.74689,1.19724,1.05666,0.205761,0.231066,0.251745,0.839504,0.566405,0.167174,0.580707,0.307606,6.33164,0.096231,0.243453,0.391844,2.14061,0.137765,0.240498,0.185392,0.243896,3.08335};

const int WLSR5N = 5;
const double WLSR5StatFix[][20] = {
{0.02549352,0.01002193,0.006005566,0.008811948,0.03859909,0.001796161,0.004312188,0.2108274,0.01335451,0.2730413,0.07862202,0.005545202,0.005058205,0.01112289,0.01296012,0.008209453,0.03210019,0.2376598,0.002668138,0.01379098},
{0.09596966,0.005026264,0.01880183,0.01582725,0.002107865,0.7215719,0.003379354,0.002257725,0.01343441,0.003013483,0.001511657,0.02805857,0.006751404,0.006454635,0.008786096,0.04798539,0.01141559,0.004934972,0.000523736,0.002188483},
{0.01726065,0.001046402,0.3627871,0.5149206,0.002075771,0.004145081,0.002563289,0.002955213,0.01558693,0.005286931,0.002693098,0.01092937,0.003006167,0.01984758,0.005467988,0.01263069,0.01082144,0.004573568,0.000253451,0.001144787},
{0.1263139,0.02095119,0.03316681,0.02790523,0.05018693,0.009007538,0.03441334,0.005855319,0.1078084,0.008061884,0.009019514,0.07050061,0.07948,0.05473468,0.09564027,0.09447839,0.09258897,0.01230413,0.01390669,0.05367769}};


	const int EFN = 12;
	const double EFStatFix[][20] =	{
	{0.072,	0.02,	0.056,	0.11,	0.012,	0.036,	0.016,	0.031,	0.095,	0.027,	0.033,	0.072,	0.027,	0.079,	0.038,	0.072,	0.102,	0.074,	0.006,	0.012},
	{0.038,	0.007,	0.015,	0.022,	0.002,	0.03,	0.012,	0.016,	0.333,	0.016,	0.015,	0.036,	0.004,	0.063,	0.299,	0.031,	0.032,	0.016,	0,	0.002},	
	{0.01,	0.012,	0,	0.002,	0.026,	0.001,	0.008,	0.129,	0.012,	0.503,	0.126,	0.002,	0.001,	0.013,	0.001,	0.002,	0.009,	0.113,	0.008,	0.01},	
	{0.034,	0.003,	0.379,	0.373,	0.001,	0.032,	0.005,	0.003,	0.01,	0.008,	0,	0.04,	0.014,	0.022,	0.004,	0.026,	0.016,	0.02,	0,	0},	
	{0.025,	0.028,	0,	0.001,	0.002,	0,	0,	0.322,	0,	0.044,	0.014,	0,	0,	0,	0,	0,	0.017,	0.537,	0,	0},	
	{0.466,	0.021,	0.011,	0,	0.001,	0.086,	0.007,	0.001,	0.019,	0.006,	0.005,	0.02,	0.001,	0.023,	0.007,	0.267,	0.035,	0.013,	0,	0.002},	
	{0.076,	0.01,	0,	0,	0.001,	0.799,	0,	0,	0,	0,	0,	0.025,	0.003,	0.001,	0,	0.049,	0.021,	0.002,	0,	0},	
	{0.083,	0.016,	0.005,	0.009,	0.001,	0.019,	0.011,	0.002,	0.016,	0.018,	0.008,	0.074,	0.016,	0.038,	0.001,	0.253,	0.408,	0.009,	0.001,	0.002},	
	{0.206,	0.006,	0.01,	0.021,	0.007,	0.028,	0.016,	0.015,	0.014,	0.015,	0.001,	0.026,	0.442,	0.023,	0.005,	0.081,	0.037,	0.032,	0,	0.005},	
	{0.001,	0.009,	0.001,	0.001,	0.403,	0.001,	0.037,	0.004,	0.001,	0.022,	0.003,	0,	0,	0,	0.003,	0.003,	0.001,	0.007,	0.006,	0.485},	
	{0.158,	0.203,	0.01,	0.001,	0.004,	0.009,	0.001,	0.031,	0.004,	0.004,	0.007,	0.017,	0.027,	0.001,	0.001,	0.028,	0.097,	0.382,	0.002,	0.002},
	{0.053,	0.058,	0.03,	0.021,	0.068,	0.019,	0.024,	0.097,	0.019,	0.077,	0.126,	0.056,	0.026,	0.059,	0.027,	0.049,	0.048,	0.05,	0.044,	0.04}};

	const double EFStatWeight[] = {0.00509464,0.00753314,0.0595801,0.0100458,0.0109109,0.0427412,0.00632106,0.0192362,0.0119168,0.0154209,0.0148037,0.014293,0.0164064,0.0272538,0.0178737,0.0125183,0.00698661,0.0140027,0.0140399,0.0109043,0.011579,0.0138661,0.0377329,0.0118586,0.0356688,0.0161327,0.0200601,0.0133087,0.0357641,0.0119353,0.0241183,0.0243098,0.0178039,0.0200988,0.00579569,0.0397441,0.0240799,0.00868054,0.0239822,0.042803,0.0335193,0.00636191,0.0547546,0.00544598,0.0354219,0.0120158,0.0047506,0.036392,0.0270346,0.00709747};

	const int UL2N = 2;
	const double UL2StatFix[][20] = {{0.089881,0.002739,0.098515,0.141889,0.007332,0.072057,0.021442,0.014026,0.092541,0.030606,0.009279,0.049615,0.07083,0.06222,0.077329,0.061715,0.055118,0.027701,0.003619,0.011549},
	{0.107894,0.017151,0.021607,0.021895,0.060355,0.025694,0.024783,0.11429,0.040631,0.158544,0.035008,0.022494,0.022624,0.020638,0.035779,0.032806,0.041926,0.125828,0.015774,0.054283}};

	const double UL2RRFix[][190] = {{2.07153,1.07358,2.05547,0.917188,1.01535,1.1908,0.557303,1.28785,1.14511,2.18181,0.83097,1.8177,2.10263,0.96588,4.90439,2.55458,4.44393,0.486419,0.589359,0.709991,0.130888,8.50171,1.68562,8.4105,4.82175,0.405685,6.38622,7.25581,3.3183,0.23296,1.68873,9.09693,15.4437,2.95104,8.84558,9.97846,21.1474,4.41566,0.0569671,1.03187,1.23146,0.0484756,0.136111,0.0411702,0.136762,5.59358,0.268012,0.513919,0.0581678,1.24126,0.47719,0.127706,0.147163,0.364746,0.110896,0.35054,0.480778,0.283534,1.08537,0.250551,0.353853,0.382293,0.235738,3.44832,0.165997,0.474658,0.601569,0.894635,0.0965316,0.222183,0.303714,5.88124,3.92836,0.148349,7.22784,6.87704,0.461813,0.177295,0.230443,0.321738,2.73572,0.675056,2.85452,8.20604,45.0917,0.715749,0.0385519,0.367292,0.118202,0.264525,2.37012,0.171682,0.378083,0.487707,2.30542,0.202452,0.238096,0.619434,0.162337,1.07222,1.32383,2.20275,1.04165,7.43619,0.591471,6.05668,4.72967,1.8489,1.18934,0.767942,2.02204,17.0088,1.20563,11.1016,16.1514,0.806118,0.0633177,0.759971,1.06076,0.616498,6.91466,32.2096,0.629908,1.95907,0.751784,2.03814,2.81367,0.264684,4.03156,3.93073,0.805351,1.70443,0.766467,0.157925,0.25722,22.5273,0.283837,0.462492,2.87769,1.70681,0.995,0.991384,5.00648,3.27355,3.05752,0.970003,0.0692198,3.53485,1.5062,1.13879,5.41849,7.00261,4.53743,3.64643,0.0785648,1.91514,0.757801,6.92589,3.24749,0.24815,0.144472,1.45427,0.483324,0.229098,1.72147,0.375537,0.336233,0.0213084,0.0910924,3.11476,1.32694,1.29559,0.910837,0.371255,0.345072,1.03759,0.825112,0.671034,1.58358,0.636157,13.6587,0.574069,1.0269,1.32899,5.06941,0.495591,0.583096,0.597489,1.64428,8.96685},
	{1.48872,0.0544995,0.169638,0.0544823,3.72237,0.0710996,0.0243062,0.0384216,0.065046,0.271017,0.110715,0.672654,0.160807,0.0665321,5.66499,1.6188,0.820632,0.0269399,0.0384389,0.10232,0.0466499,0.41274,0.843813,0.212553,0.0867758,0.0410267,0.238041,0.410709,0.648956,0.171813,0.157956,0.261716,3.66096,1.65433,0.892064,0.213299,0.359497,0.882499,0.0220052,0.228602,0.58383,0.0132892,0.0358625,0.0113612,0.0349559,2.96667,0.0405734,0.246671,0.0889505,0.452464,0.109114,0.0399537,0.0198535,0.075667,0.0170074,0.299065,0.295364,0.0507067,0.529709,0.0561693,0.181981,0.269864,0.0833387,3.76892,0.427785,0.351848,0.466517,0.214131,0.0896218,0.0682133,0.0328271,0.393237,0.291083,0.0171624,0.926716,0.717589,0.0689478,0.0599678,0.0638697,0.0375208,0.160406,0.0790409,0.162162,1.20145,4.17696,0.0538224,0.0050437,0.0595604,0.01625,0.0595891,0.269347,0.113435,0.194375,0.177132,2.53332,0.0417784,0.0388864,0.0657517,0.0134671,0.0372683,0.159143,0.116045,0.110181,3.64365,0.139996,3.82951,0.731997,0.370204,0.130597,0.0331083,0.334187,2.4548,0.0510969,1.40646,1.70796,0.106646,0.0466499,0.0663083,0.0298663,0.0377273,0.470975,4.02057,0.0251439,0.0446818,0.0552397,0.260913,0.486238,0.0519691,1.69962,8.55244,0.202081,0.356473,0.046604,0.016009,0.0405677,2.8693,0.0556012,0.193164,0.351946,0.106348,0.0518199,0.153348,0.606943,0.179519,0.0627221,0.22089,0.101172,1.15598,0.117474,0.198466,0.912629,0.648755,0.215135,0.0884972,0.0682076,0.902421,0.400931,1.75803,0.783897,0.0711168,0.0413652,0.294979,0.190875,0.0817722,0.3575,0.654528,0.263208,0.0721668,0.0439072,1.7743,0.674422,0.524293,0.139956,0.198552,0.145791,0.413343,0.25607,0.0530249,0.162661,0.0904251,2.55326,0.0494788,0.0817952,0.12343,1.12305,0.0373085,0.076631,0.0417956,0.0413021,1.28031}};

	const int UL3N = 3;
	const double UL3StatFix[][20] = {{0.104307,0.003218,0.085643,0.163928,0.016126,0.024243,0.022216,0.016012,0.105577,0.038591,0.011434,0.043722,0.018057,0.074342,0.092553,0.061232,0.061373,0.032465,0.004086,0.020876},
	{0.044015,0.00326,0.102405,0.041364,0.036695,0.168549,0.015843,0.064277,0.036061,0.118096,0.027358,0.056258,0.124914,0.018409,0.021591,0.047807,0.022696,0.025892,0.005866,0.018644},
	{0.134055,0.021008,0.020801,0.028587,0.038984,0.029069,0.028138,0.105826,0.049733,0.117458,0.02612,0.02073,0.016306,0.024509,0.044392,0.038369,0.055334,0.143398,0.017276,0.039905}};

	const double UL3RRFix[][190] = {{1.67964,1.40702,1.73693,0.141049,5.78099,1.14434,0.567951,1.16746,0.978302,1.63066,1.27535,1.0788,1.80506,0.847991,4.7159,2.26288,3.44849,0.407797,0.149284,0.0948672,0.103876,3.53891,4.32817,10.9913,4.11892,0.54211,5.41903,5.88119,3.72319,0.419903,1.5927,9.30937,16.1686,4.05211,14.0877,8.86593,12.3116,5.91663,0.0302389,2.34152,1.07811,0.102938,0.174895,0.0747079,0.236279,3.3125,1.15003,0.687326,0.0759926,1.02063,0.391772,0.167154,0.25937,0.147324,0.01535,1.27208,0.485816,0.256471,0.953909,0.179688,0.299425,0.606722,0.413101,3.10509,0.139863,0.56665,0.525624,0.676769,0.0752514,0.0341094,0.22363,1.51847,1.65782,0.0242438,4.08349,2.3369,0.162888,0.0795501,0.0204557,0.0209498,0.64704,0.0891356,0.939482,5.39532,78.9458,1.00174,0.137245,0.601764,0.379946,1.1644,2.89811,1.17767,1.22499,1.31276,11.1311,1.00119,0.757521,1.30141,0.0273731,1.49101,1.14679,2.62812,1.30034,6.94077,0.542999,6.0356,4.27592,1.6424,1.19923,0.935529,1.83904,6.14102,1.15603,8.30488,13.4049,1.59008,0.280402,0.745185,0.89554,0.682879,6.10279,22.9042,0.400303,0.666739,0.656412,1.87392,2.73827,0.309307,3.73676,3.31531,0.914563,1.57114,0.702989,0.115636,0.0492288,22.0534,0.579579,1.36222,2.34128,1.41299,1.03049,0.67914,2.92853,2.8628,1.01776,1.49949,0.293973,2.81439,1.24202,1.06663,4.02859,4.55127,5.64569,1.57526,0.198282,2.1354,0.96086,6.19539,4.46521,0.505761,0.238041,0.537251,0.582758,0.247132,1.53031,0.460666,0.512678,0.0124843,0.0830417,2.79356,1.33506,1.15532,0.779064,0.329318,0.0305683,1.08325,0.646332,0.532525,0.847793,0.0773266,16.3629,0.641539,0.961832,0.298403,4.80175,0.338821,0.115471,0.415324,0.442252,8.63433},
	{3.19771,0.463983,3.00417,0.0997895,0.439311,0.764671,0.0172366,1.22959,0.105975,0.334115,0.129404,3.81468,2.40518,1.20023,4.40014,4.48604,7.95692,0.197775,0.212484,0.750355,0.0849752,3.78973,0.995873,1.42762,0.441495,0.0256969,1.33648,2.73175,1.26453,0.44596,0.558008,1.99638,3.84892,1.46292,4.03233,2.59526,4.46217,2.93207,0.000414239,0.345693,1.3447,0.00154462,0.195394,0.000147441,0.00104613,3.13936,0.0191463,0.554989,0.127831,0.99133,1.09662,0.00691568,0.0120902,0.047476,0.00482343,0.451626,0.539325,0.00586956,1.92663,0.0106228,0.0288142,0.41075,0.596385,4.81641,0.252096,0.418374,0.849913,0.290297,0.0481079,0.16347,0.0109457,0.29828,0.705849,0.015039,3.83226,1.24908,0.00456365,0.00224672,0.0450678,0.0359826,0.139781,0.105849,1.27162,1.73664,5.69255,0.367339,0.00174823,0.430198,0.00719652,0.0170049,0.858275,0.016373,0.5589,0.740175,0.662586,0.153044,0.0845258,0.156161,0.0417679,0.0328091,1.18408,0.147947,0.16637,6.29477,0.428899,7.76406,4.49077,1.25439,0.837058,0.0740154,0.349456,3.17736,0.029755,3.55699,4.0388,0.0103138,0.000126378,0.00612231,0.0164502,0.0304571,0.361897,6.41083,0.140301,0.401468,0.0167521,0.161413,2.55691,0.424188,5.39045,9.11445,0.571432,1.4189,0.11605,0.001004,0.00468301,8.43907,0.00176227,0.0264762,0.269894,0.0986591,0.0464018,0.140287,1.87809,1.29846,0.838715,0.0443306,0.000996982,0.529896,0.255972,0.0376045,0.86927,3.10186,0.979324,0.328365,0.00885348,2.0139,0.604122,4.2028,1.43551,0.0376817,7.021e-06,0.138251,1.23976,0.738125,1.20975,1.21293,0.346121,0.0115144,0.0172436,4.27745,1.35082,1.45343,0.207941,0.102865,0.404648,1.22868,1.28412,0.209619,0.828976,0.333596,6.93848,0.0300499,0.202591,0.276901,2.30582,0.133603,0.310181,0.592769,1.63426,2.85602},
	{1.38736,0.0452016,0.119304,0.102749,3.62616,0.0625162,0.0240914,0.0283769,0.101044,0.413874,0.0820874,0.260948,0.116641,0.0439042,5.7555,1.33351,0.762457,0.0216829,0.0733438,0.010041,0.0316066,0.499922,0.614962,0.145991,0.0912519,0.0273418,0.309462,0.482746,0.778219,0.101017,0.117655,0.208652,3.86575,1.39999,0.883901,0.191592,0.51968,1.29888,0.0616121,0.133044,0.216339,0.0159137,0.0343946,0.0194401,0.0409919,8.34493,0.026417,0.288172,0.0984635,0.31991,0.0642621,0.0581202,0.00572783,0.279346,0.0700521,0.134714,0.322146,0.0521164,0.448558,0.100762,0.270222,0.313602,0.104736,3.24761,0.313195,0.286571,0.412148,0.236414,0.0842198,0.18908,0.0703074,1.04073,0.478529,0.0427034,0.489923,1.30225,0.28212,0.235462,0.180054,0.10023,0.544296,0.176058,0.282679,2.1236,3.62424,0.0666154,0.00271899,0.0477549,0.0115385,0.0615638,0.251811,0.114846,0.102445,0.120367,1.74734,0.0134915,0.0327867,0.038128,0.0386042,0.0450359,0.187024,0.165638,0.142292,3.81198,0.266448,3.26757,0.868725,0.496893,0.145646,0.0267552,0.290104,5.08417,0.0654629,1.64733,2.08418,0.153506,0.128441,0.0845856,0.0449738,0.0468302,0.600622,5.80783,0.0151615,0.0653042,0.0970971,0.356616,0.540873,0.126095,1.26161,7.64792,0.193021,0.318191,0.0376312,0.019171,0.102632,1.93706,0.0939433,0.520052,0.549037,0.151132,0.0840404,0.26339,1.08868,0.0878842,0.0874702,0.411051,0.289932,1.90837,0.159103,0.432058,1.47091,1.06136,0.0959929,0.255151,0.125543,0.866779,0.364842,2.07773,0.737717,0.0760145,0.0590657,1.00291,0.35271,0.155652,0.614727,0.417214,0.213262,0.112693,0.160483,1.39827,0.679804,0.440187,0.121016,0.191786,0.327811,0.341986,0.247594,0.047817,0.261368,0.253122,2.38146,0.0651523,0.122224,0.46706,1.0237,0.0432348,0.171027,0.0415095,0.0528548,2.00854}};


const int C10N = 10;
const double C10StatFix[10][20] = {{0.408257,0.0349388,0.00698709,0.00978467,0.00616043,0.122161,0.00391518,0.0125784,0.00596702,0.0158339,0.00813132,0.00962854,0.0394156,0.00752797,0.0081783,0.168245,0.0658133,0.0604427,0.00187516,0.00415797},
{0.102776,0.0149663,0.0155944,0.0419667,0.0180729,0.0138806,0.0158865,0.106608,0.0436344,0.113194,0.04378,0.0213272,0.0223251,0.0440685,0.0418664,0.0529608,0.108174,0.160665,0.00451472,0.0137374},
{0.0351766,0.00787065,0.000676874,0.00196868,0.0126221,0.00224206,0.00128783,0.351582,0.00188565,0.127818,0.0242632,0.00165915,0.00297716,0.00165596,0.00196786,0.00499981,0.0255378,0.388864,0.00119078,0.00375393},
{0.0408514,0.00376029,0.233381,0.0901239,0.00251082,0.115833,0.0373197,0.00255236,0.0485017,0.00521646,0.00225718,0.218565,0.0108334,0.0380451,0.0269887,0.0804527,0.030288,0.00444811,0.00108153,0.00698909},
{0.0185493,0.00704165,0.000977506,0.00248916,0.073333,0.00289529,0.0040104,0.163242,0.00435709,0.444308,0.120282,0.00248957,0.00488276,0.00835394,0.00623624,0.00516424,0.0131807,0.0968581,0.00687598,0.0144734},
{0.110675,0.00148349,0.163644,0.263846,0.00232568,0.0325228,0.0163804,0.00683349,0.0677158,0.014068,0.00489881,0.0405186,0.0298982,0.0877962,0.035219,0.0562888,0.0426922,0.0181079,0.0010339,0.00405223},
{0.0522658,0.0143325,0.0297745,0.0388387,0.0624033,0.0228101,0.155164,0.0187406,0.0439469,0.065378,0.0207189,0.0714837,0.0145475,0.073654,0.0668295,0.0549018,0.037014,0.0267512,0.0193757,0.111069},
{0.0116587,0.0105341,0.00217425,0.00242511,0.365099,0.00347091,0.0366787,0.0187185,0.00266947,0.067649,0.0143535,0.00640111,0.00311599,0.00402037,0.00509901,0.00948485,0.00737139,0.0206341,0.0509565,0.357486},
{0.0627196,0.00526629,0.0236193,0.0686285,0.00391818,0.0256175,0.0332612,0.0128968,0.227084,0.0305628,0.0124037,0.0428629,0.0140441,0.109811,0.203878,0.0483152,0.0463378,0.0197063,0.00251435,0.00655211},
{0.114552,0.00985495,0.0416192,0.0364908,0.0046606,0.0503818,0.0165233,0.00929495,0.0423027,0.0139154,0.00822408,0.0750615,0.0379222,0.0339625,0.0324009,0.261065,0.184583,0.0195769,0.0017549,0.00585383},
};

const double C10StatWeight[10] = {0.119134,0.0874372,0.103711,0.0922585,0.107049,0.132995,0.0538028,0.0691986,0.131994,0.10242};

const int C20N = 20;
const double C20StatFix[20][20] = {{0.0862413,0.0130505,0.0329909,0.0184527,0.00441553,0.0366905,0.0108013,0.00979071,0.0220195,0.0112826,0.00878215,0.0791293,0.0189273,0.0169047,0.0171944,0.317815,0.27117,0.0179753,0.00153173,0.00483429},
{0.203558,0.0348667,0.00316561,0.00708594,0.0112429,0.0195236,0.0024392,0.115257,0.00423808,0.0789777,0.0309187,0.00770524,0.0164189,0.00640441,0.00509808,0.0496777,0.111895,0.284906,0.00177626,0.00484482},
{0.0211547,0.00481886,0.000549287,0.00145396,0.0128252,0.00114309,0.00113464,0.392846,0.00135799,0.125064,0.0209789,0.0012755,0.00202472,0.00123288,0.00149462,0.00262407,0.0171914,0.386068,0.00115911,0.0036028},
{0.0376904,0.00640738,0.0109469,0.0358365,0.00363498,0.0191107,0.0329514,0.0101712,0.289763,0.0237496,0.00965289,0.0365411,0.0105337,0.0893564,0.28852,0.0356314,0.0355927,0.0144622,0.00279252,0.00665572},
{0.00845978,0.0084909,0.00244879,0.00250555,0.342046,0.00242771,0.0433214,0.0097713,0.0026741,0.0380507,0.00807248,0.00725259,0.00214187,0.00427815,0.00535899,0.00804189,0.00553221,0.012141,0.049484,0.4375},
{0.17599,0.00175587,0.130126,0.218217,0.0025277,0.0409535,0.0130708,0.00856221,0.0542946,0.0159531,0.00540458,0.0332846,0.037102,0.0707184,0.0290429,0.0793481,0.0540083,0.0249553,0.00105921,0.00362591},
{0.16344,0.00886599,0.0374273,0.0220612,0.00306413,0.529672,0.00900061,0.00175694,0.0167118,0.00611563,0.00293908,0.0438702,0.0126458,0.0137555,0.0195541,0.0829343,0.0142836,0.00579857,0.00286407,0.00323983},
{0.0917469,0.0284015,0.0133819,0.0196876,0.0998479,0.0249899,0.0449766,0.0583556,0.0164916,0.115501,0.0395995,0.0290699,0.0209916,0.0255085,0.0265853,0.0736483,0.0661518,0.0831856,0.0246464,0.0972327},
{0.0646701,0.00771176,0.0168734,0.0544978,0.0219148,0.0148894,0.0313852,0.0505983,0.0907931,0.184428,0.077484,0.0228907,0.0105004,0.0996415,0.0988016,0.0321196,0.0411766,0.0505824,0.0084303,0.0206106},
{0.0135994,0.010009,0.00079517,0.00180118,0.264097,0.00267946,0.00724019,0.0814027,0.00251581,0.366142,0.0734965,0.00184694,0.00389941,0.00464208,0.00434084,0.00436688,0.00752485,0.0573473,0.0261565,0.0660971},
{0.147804,0.00488258,0.0534743,0.0727246,0.00299039,0.0907726,0.0262289,0.00357811,0.105166,0.0126777,0.00596218,0.072663,0.0156558,0.0757166,0.0842845,0.14599,0.0634877,0.00927198,0.00159285,0.00507607},
{0.0186377,0.00549689,0.00083297,0.00202485,0.0385383,0.00217135,0.0023666,0.202081,0.00291207,0.437038,0.124186,0.00198652,0.00406723,0.00658901,0.00420552,0.00461774,0.0149904,0.118938,0.00268717,0.00563241},
{0.0477624,0.00757917,0.0141349,0.0462688,0.0130691,0.00523279,0.0165352,0.17415,0.0577575,0.112125,0.0330288,0.0209574,0.0124375,0.0429297,0.0505743,0.0264989,0.0951755,0.20937,0.00316605,0.0112466},
{0.416419,0.0406938,0.00451317,0.00632298,0.00484384,0.0946185,0.00310574,0.00764432,0.00389418,0.00998854,0.00693232,0.00917014,0.0187841,0.00613205,0.00561008,0.236077,0.0746275,0.0459225,0.00121726,0.00348258},
{0.0402296,0.0124783,0.0365524,0.0372197,0.0459095,0.0233618,0.210831,0.00934787,0.0482411,0.0360561,0.010029,0.103665,0.0098504,0.0826558,0.0735203,0.0533383,0.0310209,0.015248,0.0140077,0.106438},
{0.0323453,0.00359763,0.24315,0.0710274,0.00244293,0.101607,0.0366225,0.00314108,0.0470129,0.00519805,0.00240287,0.252045,0.00948378,0.0330831,0.0236283,0.0848355,0.0359083,0.00487046,0.000873093,0.00672477},
{0.147626,0.00323272,0.0403052,0.0576893,0.00471772,0.0330851,0.0146393,0.0108267,0.0451351,0.0256201,0.00586514,0.0211973,0.347371,0.0371554,0.0334507,0.0892065,0.0485899,0.0282336,0.00163587,0.00441772},
{0.103145,0.00617625,0.0386402,0.0923369,0.00676664,0.0202338,0.0246762,0.0376904,0.0921699,0.0376284,0.0161883,0.0435172,0.0128302,0.0786603,0.0717748,0.095145,0.137857,0.0740454,0.00221447,0.00830416},
{0.0837543,0.00207351,0.0804871,0.194776,0.00230634,0.022903,0.0268459,0.00740798,0.145929,0.019025,0.00673952,0.0518811,0.0085616,0.14565,0.0899383,0.045574,0.0451081,0.0150303,0.00107713,0.00493253},
{0.0578736,0.00111308,0.294674,0.34021,0.00170349,0.0293911,0.0139817,0.00305257,0.0363365,0.00626119,0.0027296,0.0491422,0.0156106,0.059825,0.0138314,0.0358045,0.0249942,0.00876742,0.000866434,0.0038313},
};

const double C20StatWeight[20] = {0.0559911,0.0514825,0.0812922,0.0721977,0.0556719,0.0331003,0.0589502,0.0263757,0.0307584,0.0376701,0.0303058,0.0808776,0.0263349,0.0579101,0.0371248,0.0586868,0.0561479,0.0349811,0.0544937,0.0596472};

const int C30N = 30;
const double C30StatFix[30][20] = {{0.110045,0.00190472,0.159541,0.109896,0.00166295,0.0684302,0.0137951,0.00262831,0.0358554,0.00733965,0.00247064,0.0640338,0.166936,0.0310187,0.0171295,0.138179,0.0568343,0.00823656,0.000466112,0.00359702},
{0.0874125,0.00498264,0.032612,0.0951701,0.00489966,0.0144043,0.0210627,0.0399884,0.11472,0.0301585,0.0126489,0.0382152,0.0137397,0.0798169,0.080632,0.087377,0.155862,0.0793881,0.00151228,0.00539745},
{0.0225477,0.00500182,0.000595928,0.00150305,0.0089216,0.0011571,0.000937432,0.394469,0.00136009,0.0889573,0.0189103,0.00130346,0.0018312,0.00114366,0.00149005,0.00283364,0.0189813,0.425056,0.000669375,0.00233037},
{0.0602158,0.000952546,0.290008,0.361087,0.00146256,0.0281926,0.0130501,0.00305162,0.0352705,0.00604019,0.00274606,0.0414988,0.0127175,0.0621611,0.0136833,0.0318109,0.022528,0.00932584,0.000794803,0.00340246},
{0.0101224,0.00859894,0.000637919,0.00112496,0.278538,0.00240852,0.00477534,0.0701153,0.00167485,0.413591,0.0744863,0.00129289,0.00404666,0.00350286,0.00283449,0.00370872,0.00523793,0.040886,0.0200223,0.0523939},
{0.133583,0.00105145,0.112578,0.209957,0.00167936,0.0207552,0.012133,0.00735265,0.0771772,0.0133278,0.00305717,0.0213892,0.18902,0.0565844,0.028479,0.0484054,0.0373318,0.0225174,0.000926699,0.00269464},
{0.0408277,0.0153918,0.00306349,0.00660109,0.157505,0.00581131,0.0245212,0.148751,0.00759232,0.16378,0.0385527,0.00804649,0.00583522,0.0102922,0.0124492,0.0151579,0.033222,0.154771,0.0264937,0.121334},
{0.246906,0.103941,0.00274183,0.00549448,0.0251776,0.0373263,0.00857523,0.0292404,0.00561231,0.0535091,0.0302246,0.016893,0.00780989,0.0103988,0.0106279,0.164235,0.123989,0.0955868,0.00531559,0.0163954},
{0.0549429,0.0099281,0.00929153,0.0417085,0.024386,0.0105564,0.0363512,0.0569585,0.115252,0.168183,0.0592328,0.0202958,0.00830554,0.0906036,0.130543,0.0283779,0.0412594,0.0592101,0.00963554,0.024978},
{0.0462773,0.0172727,0.0182504,0.0224266,0.133632,0.0160971,0.135785,0.0164967,0.0239396,0.0598936,0.0164507,0.0412365,0.0117413,0.0348991,0.0362984,0.0454156,0.0304388,0.0253979,0.0330338,0.235016},
{0.047438,0.00823324,0.0112117,0.0388101,0.0116644,0.00559986,0.0149157,0.183217,0.0467851,0.110069,0.0356444,0.0222454,0.0100245,0.0374051,0.041018,0.0317171,0.111435,0.219931,0.00266856,0.00996601},
{0.0213608,0.00610249,0.00129412,0.00362973,0.0371808,0.0030017,0.00384259,0.130947,0.00545678,0.456699,0.194784,0.0039879,0.00407473,0.0139566,0.00699762,0.00769915,0.0198019,0.0693208,0.00340864,0.00645457},
{0.0919632,0.0117031,0.0306717,0.0171908,0.00415894,0.0370685,0.0100793,0.00931237,0.0205386,0.0097241,0.00757673,0.0764682,0.0179686,0.016006,0.0160005,0.325447,0.274438,0.0178218,0.00138874,0.00447397},
{0.464925,0.0233329,0.00507317,0.00579942,0.00255882,0.149524,0.00232984,0.00433612,0.00285254,0.00559955,0.00393132,0.00753048,0.0186467,0.00435713,0.00430132,0.215019,0.047703,0.0292668,0.00090381,0.00200872},
{0.205133,0.00209159,0.107098,0.198973,0.00182351,0.0487574,0.0127143,0.00581247,0.0667787,0.0133472,0.00437834,0.0339418,0.0110998,0.0822742,0.0439661,0.0873962,0.0519782,0.0193174,0.00073616,0.00238214},
{0.026369,0.00394624,0.311772,0.0694354,0.00233519,0.0799842,0.0309615,0.00275212,0.0288631,0.00385797,0.00183637,0.272716,0.00624576,0.0218979,0.0133614,0.0798729,0.0324143,0.00427053,0.000722966,0.00638577},
{0.15265,0.00402153,0.0237392,0.0377949,0.00585844,0.0306214,0.014193,0.0123318,0.0375543,0.0290062,0.00644732,0.016823,0.38645,0.0341734,0.0332784,0.0880336,0.0489543,0.0313553,0.00182526,0.00488778},
{0.00802476,0.00234121,0.000338596,0.000781895,0.0311968,0.000540383,0.00100264,0.317706,0.00125983,0.373789,0.0488212,0.00063274,0.00206875,0.00155079,0.00174086,0.00120951,0.00656968,0.19446,0.00163092,0.00433436},
{0.059995,0.0016885,0.0889731,0.0962701,0.00118007,0.0588796,0.0327277,0.00214673,0.182586,0.00708764,0.00339794,0.133492,0.0141408,0.0864856,0.100054,0.0779002,0.0448817,0.00422414,0.000624903,0.00326411},
{0.0393521,0.00990719,0.0431198,0.039464,0.0188681,0.0282866,0.209572,0.00761014,0.0596434,0.025834,0.00845867,0.142548,0.00968405,0.101979,0.0838171,0.0624999,0.0347088,0.0127686,0.00546458,0.0564145},
{0.00727155,0.00678525,0.00276038,0.00250694,0.127148,0.00808347,0.0113218,0.00566096,0.00177351,0.03942,0.00798661,0.00197568,0.00410981,0.00433393,0.0140999,0.00522444,0.00430223,0.00942264,0.627357,0.108456},
{0.090707,0.00636999,0.0745873,0.0386802,0.00288592,0.475005,0.016809,0.00147211,0.0343856,0.00551498,0.00246921,0.0860678,0.0112151,0.0259377,0.0290062,0.073111,0.0153706,0.00392819,0.00229148,0.00418607},
{0.00552919,0.00721059,0.00114134,0.000968304,0.374409,0.0016071,0.0325831,0.00829184,0.00138494,0.0353678,0.00744868,0.00460866,0.00133746,0.00226922,0.00246263,0.00574027,0.00372796,0.00988026,0.0330334,0.460998},
{0.244326,0.027388,0.00315909,0.00767156,0.00606194,0.019609,0.00201894,0.101744,0.00454923,0.0468424,0.0201286,0.00624227,0.0185219,0.0053594,0.00453866,0.0497754,0.11708,0.310493,0.000957725,0.00353337},
{0.0863111,0.00244198,0.0600903,0.20361,0.00293931,0.0175221,0.0245245,0.0105994,0.148579,0.027121,0.00958244,0.0313963,0.00682768,0.167235,0.0984812,0.03478,0.0408211,0.0202271,0.00140013,0.00551054},
{0.0643926,0.00231659,0.162821,0.17627,0.00430667,0.0297139,0.0303504,0.0088163,0.072714,0.0148017,0.00567484,0.103121,0.00992703,0.0752535,0.0369049,0.0926434,0.083313,0.0161551,0.00112371,0.00938015},
{0.173668,0.00835966,0.0285985,0.0483894,0.0058274,0.0781901,0.0266135,0.00686419,0.0964012,0.0219499,0.0112303,0.0520405,0.0169661,0.0722447,0.0943629,0.15478,0.0751702,0.0172524,0.00287745,0.00821304},
{0.0347857,0.00627549,0.00923552,0.0323568,0.00343035,0.0170288,0.0306439,0.00919323,0.302085,0.0224429,0.00937208,0.0314157,0.0104447,0.0861073,0.307598,0.0326883,0.0328713,0.0130832,0.00252449,0.00641713},
{0.108774,0.0247877,0.00158232,0.00417699,0.0316523,0.0134703,0.00247658,0.164346,0.00270004,0.233715,0.0539213,0.00326798,0.0154887,0.0057932,0.0051781,0.0188188,0.0474912,0.251283,0.00376565,0.00731064},
{0.110101,0.00726998,0.0579269,0.0828181,0.0269154,0.0314463,0.0308557,0.0530866,0.029386,0.109679,0.0458729,0.0435099,0.0296431,0.0615197,0.0324325,0.0715888,0.0685882,0.0754042,0.00623241,0.0257238},
};

const double C30StatWeight[30] = {0.00957833,0.0248476,0.0636309,0.0537939,0.0295886,0.0117588,0.0132013,0.0236869,0.0261688,0.0239822,0.0257101,0.0465072,0.0546795,0.0536085,0.0270623,0.0403914,0.0474213,0.0458816,0.0214037,0.0290386,0.0123392,0.056935,0.0419688,0.0339027,0.0388777,0.0196344,0.0233086,0.0622723,0.0184803,0.0203395};

const int C40N = 40;
const double C40StatFix[40][20] = {{0.066026,0.00565867,0.105447,0.0440361,0.00131048,0.0711239,0.0168195,0.00390887,0.036669,0.0055316,0.00374124,0.159982,0.0176359,0.0273928,0.0231862,0.249769,0.150708,0.0065529,0.000672321,0.00382902},
{0.0232377,0.00379875,0.353209,0.0739378,0.00240321,0.0576668,0.0315867,0.00310928,0.0259363,0.00387116,0.00173556,0.275965,0.00631169,0.0197339,0.0122683,0.0657068,0.0270484,0.00475317,0.000760289,0.00696025},
{0.0166487,0.00366657,0.000565145,0.00133563,0.00827757,0.000889475,0.000823185,0.412937,0.00119041,0.0884689,0.0186055,0.00126222,0.001403,0.00106698,0.00125948,0.00213394,0.0162167,0.420686,0.000608205,0.00195532},
{0.239474,0.0283812,0.00447417,0.010553,0.00559911,0.013511,0.00389298,0.0765957,0.0071093,0.0358495,0.0199496,0.0120537,0.0114266,0.00865589,0.00729013,0.0847799,0.179728,0.245468,0.0009838,0.00422407},
{0.119461,0.0150527,0.0134273,0.0192173,0.0550467,0.0337676,0.0214746,0.0579002,0.0147261,0.144631,0.0561243,0.0294552,0.0631355,0.0301538,0.0233256,0.0925267,0.083123,0.0811758,0.0131636,0.0331118},
{0.0567044,0.00089248,0.29555,0.379515,0.00129723,0.023047,0.0118361,0.0031182,0.0314206,0.00601375,0.00285841,0.0364734,0.0124746,0.0609517,0.0117359,0.0300335,0.0227051,0.00946396,0.000773876,0.00313438},
{0.0179027,0.016076,0.000887041,0.00231821,0.334486,0.00398298,0.0127293,0.0404651,0.00279947,0.167614,0.0424172,0.00356977,0.00201151,0.00453955,0.00409671,0.00758416,0.00682273,0.0326045,0.0518381,0.245254},
{0.271217,0.200383,0.0021017,0.002323,0.020299,0.0502501,0.0053728,0.0150685,0.00206463,0.0330003,0.0154811,0.0141045,0.0045351,0.00482641,0.00564808,0.17642,0.0839578,0.0741934,0.00462652,0.0141271},
{0.0894737,0.00455383,0.0272183,0.127508,0.00565902,0.0115686,0.0215746,0.0469424,0.138205,0.0512035,0.0147657,0.0190192,0.00955465,0.116809,0.104003,0.0383954,0.0836653,0.0819556,0.00170794,0.00621813},
{0.0495441,0.0182506,0.0143641,0.0215379,0.141805,0.01402,0.110854,0.0247066,0.0258142,0.0700288,0.0188272,0.0315864,0.0112101,0.0316504,0.0375346,0.0456094,0.0361428,0.0369178,0.0371985,0.222397},
{0.170431,0.000974733,0.109856,0.253646,0.00133213,0.0249846,0.010139,0.00587494,0.0903324,0.0116526,0.00365127,0.0271109,0.0293614,0.09173,0.0415784,0.0561766,0.0479046,0.02033,0.000669682,0.00226373},
{0.0162725,0.00506141,0.00101821,0.00251413,0.0376246,0.00219354,0.00299143,0.132817,0.00401204,0.490444,0.192993,0.00218762,0.00343332,0.0104414,0.00548261,0.00401221,0.0127074,0.064772,0.00321076,0.00581006},
{0.0823766,0.00656943,0.0311745,0.0675531,0.00647179,0.0178962,0.0251144,0.0291162,0.0982302,0.0287904,0.0168023,0.059839,0.0114045,0.0686451,0.0734226,0.1303,0.182037,0.0540271,0.00227246,0.00795733},
{0.359497,0.0251417,0.00314844,0.00649627,0.00920205,0.119468,0.00229704,0.0458767,0.00501688,0.0468054,0.0215569,0.00334215,0.0443916,0.00490143,0.00724072,0.0465271,0.0477755,0.194216,0.00245402,0.00464504},
{0.201558,0.00323653,0.095415,0.153491,0.00256433,0.0667292,0.0155219,0.00677408,0.0547323,0.0165114,0.0060163,0.0425386,0.00919706,0.0772011,0.0430162,0.118598,0.0625473,0.0202798,0.000956551,0.003115},
{0.104273,0.00224501,0.242407,0.177482,0.00125703,0.169782,0.0132649,0.00189295,0.0220652,0.00425426,0.00164412,0.0621646,0.0317042,0.0356499,0.0147062,0.0778636,0.0288516,0.00602502,0.00069309,0.00177419},
{0.0781183,0.0194449,0.00415417,0.0116634,0.0262794,0.0111524,0.00635894,0.135453,0.00937298,0.245757,0.108778,0.015927,0.0055294,0.0240152,0.0111498,0.0408519,0.0860514,0.148276,0.00315476,0.00851085},
{0.0856592,0.0136073,0.0135062,0.00786026,0.0047153,0.0245401,0.00553791,0.0100592,0.0127319,0.0103344,0.00806758,0.0441923,0.0175274,0.00925906,0.0101233,0.340648,0.357329,0.019367,0.00142431,0.00350998},
{0.0674595,0.00216342,0.0662588,0.0865501,0.00182127,0.0368557,0.0381149,0.00332388,0.189974,0.009384,0.00394874,0.116311,0.0151208,0.093936,0.116173,0.0842204,0.0565954,0.00645142,0.00071873,0.00461894},
{0.0572262,0.00153015,0.179393,0.199226,0.00137018,0.0316472,0.0291392,0.00458046,0.101562,0.010074,0.00402046,0.108388,0.00636741,0.0903669,0.0494724,0.0621143,0.0496102,0.00859413,0.000666929,0.00464976},
{0.00360202,0.00454848,0.00208716,0.00178577,0.0855715,0.00563916,0.00649688,0.00292929,0.00104198,0.0232635,0.00445923,0.00134555,0.0024992,0.00327181,0.0102713,0.00306718,0.00259003,0.00586684,0.761782,0.067881},
{0.203202,0.00981316,0.0135012,0.00838182,0.00196196,0.618489,0.00277479,0.00118285,0.00445989,0.00398268,0.00206318,0.0143744,0.00858704,0.00445146,0.00838957,0.073992,0.0108922,0.00607691,0.00186061,0.00156387},
{0.00508988,0.00617765,0.00161262,0.00120404,0.356313,0.00163342,0.0393461,0.00590888,0.00137137,0.0249344,0.00497952,0.0057093,0.00141364,0.00246931,0.00287408,0.00595277,0.00365368,0.00819341,0.0357987,0.485365},
{0.0403336,0.00815495,0.00982186,0.0375407,0.0119141,0.00479344,0.0176736,0.189342,0.0607377,0.105186,0.03056,0.0216052,0.00775506,0.0383639,0.0540186,0.025711,0.100991,0.221091,0.002878,0.0115277},
{0.0790086,0.00249479,0.0546012,0.199788,0.00237734,0.0192656,0.02707,0.00756675,0.155311,0.0254542,0.00980244,0.0309384,0.00566407,0.184338,0.106544,0.0332371,0.0359575,0.0145306,0.00116828,0.00488208},
{0.0722241,0.00647553,0.119488,0.134589,0.0348233,0.0287815,0.0699011,0.0173589,0.0490342,0.051987,0.0154411,0.067893,0.0145597,0.070897,0.0489728,0.058958,0.0425973,0.0317884,0.00879138,0.0554387},
{0.108584,0.00125026,0.152967,0.166485,0.00145535,0.0336098,0.0134902,0.00388218,0.0576227,0.00898614,0.0024339,0.0441956,0.19901,0.0405398,0.020645,0.084675,0.0454715,0.0113416,0.000590283,0.00276502},
{0.0309526,0.0065594,0.00823521,0.0291974,0.00368916,0.0154206,0.0310385,0.00982516,0.306263,0.02379,0.00970717,0.0301337,0.00950291,0.0832608,0.319589,0.0295285,0.0303052,0.0133037,0.00281253,0.00688506},
{0.00989537,0.00282767,0.000374823,0.00091821,0.0298607,0.000699707,0.00104195,0.311504,0.00139605,0.375039,0.0474451,0.000730793,0.00252963,0.0017337,0.00196045,0.0014628,0.0075739,0.1973,0.00167998,0.00402599},
{0.116321,0.00347923,0.0731918,0.138088,0.00941177,0.0193193,0.0160241,0.0712243,0.035512,0.0771474,0.0242841,0.0250164,0.0508927,0.0586677,0.0273321,0.047556,0.0726552,0.123571,0.00268927,0.0076166},
{0.128522,0.0172929,0.040275,0.0250692,0.0199118,0.112703,0.0606981,0.010935,0.028875,0.0258416,0.0167593,0.117984,0.0180675,0.0439706,0.0373073,0.174149,0.0648968,0.0182067,0.0063575,0.0321772},
{0.0372287,0.014494,0.00237032,0.00485851,0.146377,0.00464339,0.0186795,0.182046,0.00581985,0.17801,0.0371334,0.00533773,0.00485386,0.00790971,0.0094528,0.0103571,0.0284162,0.184992,0.0211294,0.0958905},
{0.0535644,0.00962562,0.0113537,0.0391699,0.0264214,0.0120279,0.0384888,0.0522748,0.0996038,0.189239,0.0712219,0.0239173,0.00837206,0.0928585,0.11598,0.0299114,0.0389485,0.0500948,0.0104232,0.026503},
{0.133242,0.0246514,0.00127392,0.00404615,0.0156995,0.00891392,0.00158647,0.197128,0.00237132,0.125129,0.0286947,0.0022705,0.0118846,0.00308435,0.00331477,0.0171462,0.0563298,0.356621,0.00173418,0.00487784},
{0.149866,0.00354374,0.0280355,0.0435381,0.00475339,0.0311113,0.0140626,0.0101953,0.0393125,0.0251434,0.00515483,0.0176453,0.39238,0.0348151,0.0326607,0.0874497,0.0473307,0.0271597,0.00152152,0.00432083},
{0.421437,0.018761,0.00733051,0.00868378,0.00271561,0.0902333,0.0030262,0.00393628,0.00515087,0.00471933,0.00383066,0.012159,0.020894,0.00727486,0.0061426,0.290119,0.0651922,0.0252211,0.000810824,0.00236228},
{0.177071,0.00783489,0.02265,0.0509767,0.00405142,0.089739,0.0220667,0.00595198,0.125769,0.020537,0.00929825,0.0311657,0.0264088,0.0752471,0.133278,0.116959,0.0565567,0.0165087,0.00299471,0.00493467},
{0.0293984,0.00317291,0.109971,0.046427,0.00150396,0.422242,0.0272495,0.000799733,0.0622314,0.00376343,0.00166571,0.148362,0.00564818,0.0388688,0.0370902,0.0472252,0.0086569,0.00203639,0.000917602,0.00276931},
{0.0265779,0.0101369,0.0280314,0.0269057,0.0276961,0.0173377,0.281513,0.0064647,0.0474749,0.026821,0.00723753,0.13186,0.0083015,0.0989711,0.0791105,0.0426277,0.0259043,0.0100147,0.00785289,0.0891598},
{0.00960965,0.00773017,0.000633186,0.00104719,0.263017,0.00202274,0.00390014,0.0733098,0.00149315,0.445169,0.0732575,0.00131044,0.00427681,0.00338994,0.00271362,0.00361174,0.00579284,0.0425173,0.0181276,0.0370698},
};

const double C40StatWeight[40] = {0.0223854,0.0338892,0.0577169,0.0252416,0.0108608,0.0462374,0.0102293,0.0147524,0.0143161,0.0182303,0.0204025,0.0425505,0.0248627,0.0105893,0.0188239,0.00866634,0.0148496,0.0343037,0.0225335,0.0174069,0.0112208,0.0443532,0.0392573,0.0196757,0.028769,0.0114441,0.0112339,0.0582694,0.0444272,0.0112011,0.0145176,0.0114629,0.0239628,0.0266266,0.0481201,0.0371147,0.0160477,0.0237249,0.0235226,0.0261998};

const int C50N = 50;
const double C50StatFix[50][20] = {{0.115682,0.00212412,0.125699,0.0919063,0.0016175,0.0391977,0.0108596,0.00344834,0.0344439,0.0125587,0.00467132,0.0530851,0.206661,0.0264814,0.0147638,0.154493,0.0841319,0.0134754,0.000392551,0.00430762},
{0.0983768,0.00160659,0.0456633,0.0935028,0.00112009,0.0251388,0.0191777,0.00367483,0.151407,0.0115555,0.00258422,0.0266318,0.229718,0.0694385,0.0855866,0.0696565,0.0498854,0.0123983,0.00069296,0.00218381},
{0.0214598,0.00446298,0.000555089,0.00135258,0.00855351,0.00112246,0.000813505,0.399631,0.00128242,0.0868331,0.0170323,0.00117488,0.00183519,0.00101872,0.00142423,0.00233511,0.0161891,0.430205,0.00061736,0.00210206},
{0.117155,0.032654,0.00620066,0.00998032,0.0657132,0.0333659,0.0134621,0.0557335,0.00794829,0.199356,0.0938018,0.0221831,0.0115415,0.0204903,0.0146873,0.0842555,0.0920314,0.0798031,0.0127314,0.0269048},
{0.0371284,0.0131607,0.00255462,0.00465054,0.156922,0.00416133,0.019977,0.178808,0.00498614,0.165159,0.0352649,0.0057612,0.00512406,0.00679862,0.00826841,0.0102949,0.0275498,0.181807,0.0221255,0.109498},
{0.0540087,0.000836323,0.306245,0.38187,0.00111469,0.0245756,0.0116353,0.00254509,0.030937,0.00536416,0.00256903,0.0359123,0.0122515,0.0586791,0.0115134,0.0277199,0.020314,0.00828799,0.000720693,0.00290032},
{0.0404572,0.028652,0.00495693,0.0068813,0.242849,0.0103569,0.0318377,0.0329142,0.0105561,0.143431,0.0485779,0.0105988,0.00489299,0.0118557,0.0126873,0.0229136,0.0212457,0.0349069,0.0593874,0.220042},
{0.319667,0.138968,0.00144847,0.00200138,0.00997903,0.0503464,0.00296588,0.0182219,0.00202126,0.0260477,0.0136422,0.00936868,0.00591365,0.00358722,0.00435407,0.1656,0.118413,0.0996664,0.00164228,0.00614589},
{0.0977328,0.00251083,0.0724816,0.197027,0.00556062,0.0117997,0.0154288,0.0563751,0.0613631,0.047368,0.0159721,0.0241106,0.0149,0.0909736,0.034432,0.0408092,0.0929289,0.111482,0.00122744,0.00551621},
{0.0195453,0.0107627,0.0127492,0.011794,0.14032,0.0061666,0.215761,0.00826718,0.0118897,0.0328258,0.00664433,0.0423533,0.00528472,0.0258824,0.024497,0.0203711,0.0124449,0.0133889,0.0330587,0.345993},
{0.185338,0.00185224,0.161345,0.234878,0.00192286,0.054278,0.0109591,0.00550122,0.0382027,0.00925008,0.00343293,0.033406,0.0233824,0.0551124,0.0201268,0.0884267,0.051789,0.0177494,0.000835994,0.00221142},
{0.0143603,0.0036047,0.000411253,0.00108887,0.0196479,0.00187385,0.000991664,0.285685,0.00213672,0.317288,0.150556,0.00155703,0.00122034,0.0028111,0.00194519,0.00344669,0.0196686,0.167294,0.000843478,0.00356936},
{0.150577,0.00957632,0.0227565,0.0358641,0.0104881,0.0285062,0.0171219,0.0577109,0.0276149,0.0531781,0.0298859,0.0514957,0.0243637,0.0354202,0.0272841,0.132633,0.169554,0.104387,0.00219421,0.00938724},
{0.277615,0.0280749,0.00194439,0.00537302,0.0070772,0.0234861,0.0012257,0.0949678,0.00308357,0.0474515,0.0189616,0.0034483,0.0134815,0.00374076,0.00354151,0.0423818,0.0932708,0.326207,0.000963513,0.00370386},
{0.205879,0.00249091,0.075426,0.161651,0.00210732,0.0463391,0.0159993,0.00515537,0.0913036,0.0166911,0.00536596,0.0366042,0.00849072,0.0973331,0.0696431,0.088244,0.0501877,0.0172747,0.000776174,0.0030381},
{0.0704473,0.00149732,0.187469,0.208434,0.0013217,0.0409359,0.0247326,0.00339884,0.0739754,0.00761491,0.00270559,0.0965448,0.00673196,0.0820493,0.0355811,0.0832004,0.0610002,0.00788473,0.000646918,0.00382803},
{0.108767,0.00735776,0.0100122,0.0249269,0.015562,0.0114116,0.00730171,0.0923916,0.0138538,0.112308,0.0251594,0.00831075,0.259979,0.0187962,0.0142911,0.0329858,0.0608007,0.162244,0.0049709,0.00857003},
{0.109878,0.0295589,0.000867374,0.00267084,0.028604,0.0101436,0.00159952,0.171472,0.0015037,0.241889,0.0488425,0.00201796,0.00631863,0.00393786,0.00336164,0.0148413,0.0393773,0.273667,0.00359303,0.00585493},
{0.058736,0.00692528,0.0144651,0.0412403,0.0154694,0.015436,0.0243546,0.0388984,0.0996366,0.207192,0.103832,0.0289715,0.00833815,0.108218,0.0976056,0.0350037,0.0432645,0.0338957,0.00652556,0.0119916},
{0.0642736,0.00525071,0.144305,0.153274,0.0286093,0.0265312,0.0545105,0.0231701,0.0452879,0.058794,0.0192273,0.071063,0.0146855,0.066931,0.0405954,0.054046,0.043602,0.0350962,0.00655203,0.0441953},
{0.00487716,0.00503139,0.00209095,0.00192293,0.0983777,0.00615619,0.00725617,0.00400789,0.000993075,0.0277171,0.00537714,0.00145717,0.0029039,0.00350461,0.0112195,0.00329284,0.00295187,0.00708726,0.724652,0.0791231},
{0.0552756,0.0115894,0.00846155,0.0432312,0.00961795,0.0074182,0.0218432,0.0843826,0.163327,0.0631707,0.0228359,0.0221225,0.00652522,0.0639538,0.146077,0.0387464,0.100713,0.118118,0.00298298,0.00960875},
{0.00449551,0.00619902,0.000890609,0.000674178,0.400884,0.0013356,0.0223834,0.00805107,0.00105813,0.0341606,0.00701178,0.00318917,0.00130624,0.00157564,0.00181707,0.00502515,0.00337803,0.00946203,0.0345278,0.452575},
{0.0505469,0.00768183,0.0101983,0.0470721,0.0160596,0.00517598,0.0176387,0.169463,0.0497501,0.209444,0.0474166,0.0132404,0.00844474,0.053753,0.051076,0.0149657,0.03846,0.172926,0.00437246,0.0123139},
{0.0749936,0.0124353,0.0203869,0.0100161,0.0038092,0.0270453,0.00581932,0.00743392,0.014732,0.00838354,0.00640172,0.0530316,0.0176075,0.0096947,0.0117746,0.3482,0.348235,0.0157071,0.00137596,0.00291601},
{0.0569785,0.00176591,0.102737,0.106157,0.00132179,0.0360979,0.034465,0.00311879,0.185459,0.00858711,0.00431374,0.139374,0.00855798,0.0898676,0.101042,0.0669467,0.0422839,0.00552428,0.000686497,0.00471478},
{0.0746973,0.00394557,0.0423454,0.0906599,0.00363883,0.0183504,0.0229406,0.0156166,0.12297,0.0167059,0.00901115,0.0672502,0.0127747,0.0811334,0.0755385,0.132886,0.170553,0.0324157,0.00152093,0.00504616},
{0.17313,0.00938612,0.0203504,0.0387504,0.00448721,0.0867075,0.0227177,0.00492238,0.125544,0.0156727,0.00787132,0.0400476,0.0180405,0.0680443,0.133335,0.142136,0.0670614,0.0136107,0.00287027,0.00531475},
{0.00751784,0.00212117,0.000341917,0.000667876,0.0310677,0.000466913,0.000898937,0.320302,0.00103903,0.384419,0.0335748,0.000602706,0.00243131,0.00137426,0.00152778,0.00115498,0.00552329,0.199295,0.00172716,0.00394622},
{0.113459,0.0281671,0.0268294,0.0283293,0.0928796,0.0494616,0.0733971,0.0152335,0.0176739,0.037425,0.0140126,0.0539481,0.0259048,0.0303861,0.0259068,0.121621,0.0621544,0.0350518,0.0240853,0.124074},
{0.0810974,0.00899714,0.0930492,0.0372909,0.002713,0.0907986,0.0275326,0.00533957,0.0362449,0.00831441,0.00615441,0.177093,0.0120848,0.0316411,0.025247,0.233557,0.106302,0.00814021,0.000950692,0.00745168},
{0.00757098,0.00716117,0.00049456,0.000891326,0.297029,0.00155605,0.00342729,0.0711141,0.00107067,0.416918,0.0635143,0.000864035,0.00349893,0.00251506,0.00186372,0.00271598,0.00424124,0.041488,0.0202374,0.0518279},
{0.086899,0.00542266,0.0274343,0.117345,0.00568342,0.0162895,0.0386176,0.0182492,0.149526,0.048245,0.0150062,0.0265186,0.00577334,0.159013,0.149227,0.0416278,0.0492401,0.0283951,0.00272523,0.00876225},
{0.0501736,0.0113806,0.00768113,0.0191841,0.0119036,0.00521331,0.00988204,0.203482,0.0206446,0.0870652,0.036001,0.0197961,0.0061624,0.0180415,0.0178215,0.0399964,0.162255,0.262373,0.00172457,0.00921833},
{0.177098,0.00357587,0.0740854,0.0713923,0.00385864,0.211235,0.0161904,0.0040472,0.0370753,0.00826202,0.00369757,0.0494737,0.0543871,0.0396583,0.0340392,0.140615,0.0532238,0.0108131,0.00212448,0.00514764},
{0.242384,0.0115188,0.00720908,0.00544261,0.00229719,0.593273,0.00205723,0.00208173,0.00287156,0.00491,0.00271204,0.00899551,0.00763059,0.00285962,0.00600369,0.0738385,0.0108329,0.00968775,0.00178961,0.00160391},
{0.032912,0.00319267,0.143899,0.0532755,0.00166462,0.463509,0.0175283,0.000876108,0.0317056,0.00297349,0.0012156,0.131603,0.00688601,0.0251373,0.0184844,0.0491059,0.00984294,0.00239178,0.0008917,0.00290456},
{0.0388306,0.00489943,0.0491655,0.0413874,0.00170354,0.345044,0.0421499,0.00103798,0.123725,0.00726049,0.00335164,0.119727,0.00562115,0.0664068,0.0881106,0.0460192,0.00791813,0.0020819,0.00195492,0.00360516},
{0.0373301,0.00916334,0.0396965,0.0362168,0.0144959,0.0269081,0.220512,0.00638711,0.0634517,0.0225147,0.00718608,0.152708,0.00995802,0.108449,0.0882553,0.0609295,0.0332848,0.0105312,0.00429315,0.047729},
{0.0171508,0.00565843,0.00111564,0.00268175,0.0490847,0.00219974,0.00370226,0.116938,0.00407925,0.500942,0.1832,0.00280387,0.00400697,0.0115255,0.00613355,0.00539456,0.0140434,0.0579479,0.00419195,0.00719895},
{0.0238117,0.00329378,0.370777,0.0831676,0.0022275,0.053433,0.0290274,0.00288609,0.0258098,0.00370314,0.00159506,0.259554,0.00668902,0.0201877,0.0117854,0.0626213,0.0278627,0.00464381,0.000688992,0.0062349},
{0.209736,0.0105937,0.0311812,0.0613839,0.00991105,0.0855144,0.0226071,0.0296058,0.0455281,0.0792591,0.0306389,0.019447,0.0213982,0.0610672,0.0588191,0.0820738,0.0527883,0.0722389,0.00513,0.0110785},
{0.0295833,0.00588894,0.00852018,0.0296344,0.00323424,0.0154495,0.02992,0.00752729,0.317117,0.0212988,0.00874154,0.0312424,0.00843308,0.0831654,0.323587,0.0294177,0.0283085,0.0100345,0.00257807,0.00631797},
{0.0937098,0.00289961,0.0255803,0.0577481,0.0140839,0.0182984,0.0300827,0.0312422,0.0867966,0.067914,0.01767,0.0206426,0.213942,0.0611154,0.0902988,0.0526724,0.0464314,0.0509513,0.00365078,0.01427},
{0.419507,0.0176715,0.00719436,0.00880545,0.00285112,0.0865515,0.00330323,0.00419601,0.00503365,0.0048203,0.00393323,0.012443,0.0192167,0.00750218,0.00609116,0.295973,0.0660748,0.0253174,0.000942289,0.00257126},
{0.0368106,0.0124259,0.0102859,0.0343205,0.0759532,0.00772969,0.0938663,0.0458547,0.0701169,0.133403,0.031916,0.0265402,0.00749485,0.0673302,0.10388,0.0271625,0.0353789,0.0485031,0.0254607,0.105568},
{0.0665479,0.00185956,0.183575,0.149215,0.00127224,0.0634659,0.019684,0.00262142,0.0655694,0.00750609,0.00151282,0.0699101,0.182105,0.0446839,0.0292035,0.075452,0.0278833,0.00484561,0.000723776,0.00236318},
{0.16002,0.00397537,0.019682,0.0286161,0.00458768,0.0326662,0.0114325,0.00722291,0.0232558,0.0202542,0.00379358,0.0142832,0.460797,0.0260458,0.0217875,0.0913633,0.0427326,0.0227118,0.0012566,0.0035159},
{0.0843584,0.00129715,0.0847585,0.260552,0.00143564,0.0191268,0.0174708,0.00556223,0.144172,0.014456,0.00550323,0.0336163,0.00724868,0.160704,0.0742581,0.0323411,0.0355032,0.0136662,0.000760206,0.00320994},
{0.152149,0.000925399,0.116471,0.220438,0.00150071,0.0172039,0.0095692,0.00747151,0.0529323,0.0148756,0.00356173,0.0157411,0.217335,0.0489135,0.0171707,0.044102,0.0323715,0.023915,0.000908902,0.00244332},
};

const double C50StatWeight[50] = {0.00509464,0.00753314,0.0595801,0.0100458,0.0109109,0.0427412,0.00632106,0.0192362,0.0119168,0.0154209,0.0148037,0.014293,0.0164064,0.0272538,0.0178737,0.0125183,0.00698661,0.0140027,0.0140399,0.0109043,0.011579,0.0138661,0.0377329,0.0118586,0.0356688,0.0161327,0.0200601,0.0133087,0.0357641,0.0119353,0.0241183,0.0243098,0.0178039,0.0200988,0.00579569,0.0397441,0.0240799,0.00868054,0.0239822,0.042803,0.0335193,0.00636191,0.0547546,0.00544598,0.0354219,0.0120158,0.0047506,0.036392,0.0270346,0.00709747};

const int C60N = 60;
const double C60StatFix[60][20] = {{0.153436,0.00473066,0.0546757,0.0529324,0.00355543,0.110378,0.017448,0.00503439,0.0603929,0.0130294,0.00755506,0.0796727,0.0249524,0.0514333,0.0444389,0.202963,0.0957668,0.0101809,0.00144445,0.00598003},
{0.0281985,0.00195035,0.00915493,0.038853,0.00159193,0.00840283,0.0155385,0.0107873,0.347604,0.0217787,0.0109905,0.0312955,0.00715399,0.0939884,0.303106,0.0197479,0.0328352,0.0135303,0.000920999,0.0025714},
{0.00836807,0.00204339,0.000222837,0.000473154,0.0123984,0.000482575,0.000518999,0.376845,0.000629617,0.260833,0.0315701,0.000612345,0.000959592,0.000949869,0.000731977,0.000974638,0.00499908,0.293808,0.000865776,0.00171323},
{0.222723,0.164029,0.000799759,0.00272808,0.0228667,0.0202703,0.00371839,0.0455406,0.00228329,0.088335,0.0348095,0.00612065,0.00354716,0.00510519,0.00648461,0.085004,0.101285,0.169858,0.00484248,0.00965},
{0.041214,0.0204716,0.00174344,0.00372634,0.2239,0.00691075,0.0180294,0.115428,0.00429003,0.169356,0.0414067,0.00510677,0.00584164,0.00575385,0.00676271,0.0149106,0.0239548,0.120034,0.0332237,0.137935},
{0.048055,0.00378018,0.208461,0.185518,0.0196001,0.0191402,0.0872162,0.00946854,0.0375741,0.0277341,0.00883084,0.0940629,0.00812678,0.0747677,0.0308438,0.0439681,0.0324589,0.0181769,0.00346657,0.03875},
{0.00628487,0.00730276,0.000532415,0.000974149,0.397455,0.00195782,0.0102225,0.0180592,0.00162059,0.117906,0.0234722,0.00303425,0.00201656,0.00344091,0.00262469,0.00569033,0.00370918,0.0152754,0.0598639,0.318557},
{0.181501,0.420563,0.00251455,0.000708814,0.00689764,0.0461854,0.00143746,0.00417455,0.000647425,0.00983105,0.00416114,0.0148485,0.00387672,0.0014097,0.00268454,0.186454,0.068719,0.0345742,0.00270835,0.0061033},
{0.0600741,0.00707198,0.0170572,0.0127487,0.114091,0.0343564,0.0305859,0.0204571,0.0212367,0.0994551,0.0318166,0.0134869,0.0297628,0.014247,0.0367643,0.0505793,0.0339368,0.0217422,0.231281,0.119249},
{0.0708395,0.0131641,0.0324483,0.0508264,0.059436,0.0183309,0.0567273,0.0650369,0.0343618,0.128226,0.0390363,0.0416822,0.0135608,0.0494875,0.0474098,0.0551343,0.064226,0.078961,0.0137118,0.0673934},
{0.0617689,0.000719984,0.343023,0.379251,0.000981419,0.0284659,0.00795921,0.00169996,0.0216077,0.00399453,0.00197343,0.0303082,0.0174791,0.0307856,0.00763329,0.0337832,0.0203427,0.00589923,0.000613027,0.00171028},
{0.042156,0.00444503,0.0164385,0.0531949,0.00867283,0.0134711,0.0177765,0.0267728,0.132374,0.196724,0.118283,0.0286981,0.00518379,0.139369,0.104207,0.0255853,0.0333292,0.0217067,0.00458523,0.00702815},
{0.281481,0.0258338,0.00643857,0.0115046,0.00504817,0.0270055,0.00546297,0.0188216,0.00987128,0.0190993,0.015872,0.0172868,0.012951,0.0133102,0.0100367,0.188681,0.24271,0.0837524,0.00122746,0.00360529},
{0.276919,0.0318293,0.00116725,0.00247525,0.00708439,0.0199647,0.000517086,0.0983109,0.00162322,0.0489264,0.0173415,0.00213153,0.00701797,0.00182163,0.0017227,0.0336349,0.0814141,0.362092,0.000711814,0.00329423},
{0.15778,0.00533181,0.064829,0.0913587,0.00250478,0.090621,0.053381,0.00291716,0.0782149,0.0156226,0.0045759,0.0570403,0.00670778,0.106537,0.111214,0.092931,0.0393123,0.00850409,0.00285758,0.00775903},
{0.0593735,0.00153322,0.218948,0.181942,0.00108383,0.0813707,0.0220478,0.00209936,0.0750172,0.00561913,0.00218717,0.115118,0.0109737,0.0688752,0.0354741,0.0726449,0.0380238,0.00426697,0.000734646,0.00266649},
{0.0978066,0.00641384,0.0120781,0.0291283,0.0198964,0.0128913,0.0128037,0.0723904,0.0245631,0.127944,0.0303267,0.0101844,0.272352,0.0307876,0.0265576,0.0350549,0.0484557,0.115203,0.00468425,0.0104774},
{0.0124023,0.00418857,0.000603732,0.0012524,0.0610615,0.00113068,0.00135421,0.234417,0.00167182,0.455056,0.0667308,0.00092391,0.00370762,0.00329574,0.00306804,0.00194209,0.00676129,0.129048,0.00389372,0.00749118},
{0.0794231,0.00199195,0.0587236,0.12469,0.001287,0.0325432,0.0238605,0.00362772,0.21476,0.0097988,0.00418462,0.0662793,0.014241,0.114388,0.129474,0.0615807,0.0477334,0.00781872,0.000652537,0.00294202},
{0.0248149,0.00278397,0.427883,0.143239,0.00146804,0.0643968,0.0185737,0.00225069,0.0179274,0.00345586,0.00157145,0.188892,0.0073768,0.0210778,0.00835529,0.0377003,0.0187768,0.00460727,0.000589186,0.00426027},
{0.00170034,0.00290594,0.00107115,0.00117317,0.0513377,0.00355796,0.00276305,0.00122912,0.00042738,0.0127421,0.00256713,0.00042229,0.00135367,0.00164242,0.00606743,0.00118717,0.00140331,0.00331373,0.864044,0.0390913},
{0.0468361,0.00599541,0.0185616,0.0705436,0.00801772,0.00454353,0.0152063,0.155061,0.0866248,0.0824253,0.0245854,0.0205604,0.00814856,0.0557031,0.0639797,0.0237026,0.0962054,0.204749,0.00183687,0.00671317},
{0.0258765,0.0142504,0.0107037,0.010565,0.187643,0.0073065,0.141108,0.0114778,0.0119179,0.0407992,0.0098799,0.0298384,0.00512288,0.0158529,0.0201097,0.02757,0.0170765,0.0187502,0.0405125,0.353639},
{0.0296285,0.00900465,0.000885102,0.00330465,0.0408798,0.00279695,0.00507015,0.25834,0.00466909,0.266809,0.0770825,0.00349444,0.00269182,0.00554811,0.00464003,0.00685381,0.0322266,0.22719,0.00355061,0.0153353},
{0.0555726,0.00974186,0.014039,0.00694432,0.00344967,0.0157957,0.00416313,0.00698265,0.0139225,0.00752712,0.00587627,0.0409064,0.0119733,0.00687277,0.00984479,0.348247,0.421366,0.0131119,0.00100615,0.00265768},
{0.0682671,0.0141433,0.0360279,0.0497415,0.0803376,0.0268697,0.112767,0.0132647,0.0482609,0.0544494,0.0170034,0.0530661,0.019195,0.0612274,0.0615207,0.067184,0.0443996,0.0267714,0.0199958,0.125507},
{0.0792619,0.00495033,0.0436646,0.0829639,0.00491311,0.0183429,0.0233169,0.0249427,0.0932578,0.0221483,0.0120893,0.0635289,0.012636,0.0666365,0.0638377,0.133485,0.191675,0.0503102,0.00180401,0.00623531},
{0.0731759,0.0123459,0.0110007,0.0349943,0.00931488,0.0224509,0.0386903,0.0246226,0.178334,0.0508307,0.0185741,0.0324201,0.0148723,0.0858952,0.210534,0.0603181,0.0649575,0.038504,0.00510464,0.0130597},
{0.0878403,0.0171147,0.00328039,0.0101466,0.0191954,0.00870909,0.00549022,0.198776,0.00820969,0.175646,0.0417233,0.00608012,0.0111284,0.0109832,0.0110332,0.0209863,0.0697719,0.28252,0.0031744,0.00819055},
{0.0990216,0.0108996,0.0118798,0.0279,0.0214442,0.024025,0.0123445,0.0776565,0.0322818,0.235451,0.120767,0.021115,0.00756555,0.0557711,0.0349352,0.052417,0.0649785,0.0771067,0.00470758,0.00773287},
{0.0601641,0.0315067,0.0337189,0.00598395,0.00695259,0.0543081,0.0531524,0.00706508,0.0139598,0.00706981,0.00882987,0.278352,0.00758343,0.0210646,0.0161995,0.28298,0.0860317,0.0100954,0.00149666,0.0134849},
{0.00497817,0.00529787,0.000531537,0.000763055,0.335112,0.00150513,0.00415752,0.0443829,0.00116151,0.441742,0.0602807,0.000701221,0.00278477,0.00240899,0.00184123,0.00257958,0.00302885,0.0237677,0.0171303,0.0458456},
{0.0251997,0.00970995,0.00410123,0.0262572,0.0429092,0.00380673,0.0431939,0.0974043,0.0824313,0.24472,0.0539323,0.0142032,0.00526585,0.0620071,0.111447,0.00960931,0.0251183,0.0783303,0.0146572,0.0456965},
{0.0230362,0.00485801,0.000626444,0.00157626,0.0072127,0.00115653,0.000889924,0.39529,0.00146635,0.0576537,0.0140987,0.00135344,0.00201779,0.000987005,0.00147487,0.00287702,0.0205581,0.460366,0.000547769,0.00195391},
{0.140878,0.00322137,0.0285077,0.0423839,0.00412311,0.0299558,0.0131321,0.00669145,0.0383782,0.019512,0.00362769,0.0171298,0.438347,0.0320632,0.0297808,0.08514,0.0422766,0.0198195,0.00132369,0.00370876},
{0.444249,0.0235405,0.0057193,0.00648899,0.00244132,0.0822688,0.00255057,0.00336151,0.00380971,0.00409901,0.00286831,0.0102305,0.0162891,0.00559976,0.00502166,0.299997,0.0559665,0.0226608,0.000773543,0.00206398},
{0.0898718,0.00441315,0.0129167,0.0058957,0.000807357,0.808756,0.00165179,0.000433928,0.0033939,0.00155645,0.000425342,0.0130068,0.00341281,0.00238065,0.00709583,0.0362877,0.00328875,0.00202896,0.00152239,0.000853745},
{0.055084,0.00358249,0.12738,0.0655885,0.0040375,0.0637608,0.0404883,0.00755742,0.0867683,0.0136305,0.00816842,0.187783,0.0110682,0.0527969,0.0472254,0.126338,0.0752544,0.0102728,0.00135637,0.011859},
{0.0117681,0.00703883,0.0144712,0.0110901,0.0338266,0.00496266,0.433719,0.0061337,0.0137929,0.0298795,0.00762376,0.0844145,0.00813461,0.103834,0.0442559,0.0140571,0.0108277,0.00837578,0.00806831,0.143725},
{0.0159286,0.00805199,0.00109669,0.00213242,0.0743651,0.00250864,0.00491921,0.0501544,0.00355994,0.530763,0.216009,0.00326926,0.00452474,0.0134552,0.00480987,0.00669222,0.0119092,0.0281012,0.00709281,0.0106565},
{0.0195973,0.0034539,0.309985,0.0250778,0.00197065,0.0627529,0.0295961,0.00326504,0.0240964,0.00282469,0.000842506,0.32891,0.00490628,0.0116197,0.0105143,0.106498,0.0438054,0.0040302,0.000633396,0.0056198},
{0.0833804,0.00819811,0.0686821,0.0227415,0.00241326,0.0982292,0.00739549,0.00174712,0.0129342,0.00396531,0.0019558,0.0969824,0.0355924,0.0121521,0.0125871,0.311561,0.211337,0.00471386,0.0016329,0.00179911},
{0.0181409,0.00851436,0.00333335,0.00965319,0.00354269,0.0224553,0.0375239,0.00350905,0.320807,0.0149147,0.00460989,0.0233205,0.00871975,0.0526694,0.412966,0.0262309,0.0131791,0.00447465,0.0034767,0.00795882},
{0.249423,0.0122863,0.0169234,0.0370491,0.00773609,0.0347468,0.00870696,0.0595718,0.0177065,0.0451065,0.0204556,0.016412,0.0686404,0.0228502,0.0185482,0.0889296,0.0986017,0.167252,0.00286039,0.00619385},
{0.141974,0.00688567,0.0537452,0.040754,0.00402861,0.398811,0.0152896,0.00166276,0.0314274,0.00923483,0.00554255,0.0576297,0.0180329,0.028624,0.0373946,0.112373,0.0242478,0.00484629,0.00259091,0.00490542},
{0.0178903,0.0290304,0.00310543,0.00405034,0.0238913,0.010048,0.12527,0.0016708,0.023334,0.0722357,0.0116142,0.0155854,0.000993842,0.105182,0.195884,0.0181676,0.018626,0.0113865,0.226055,0.0859787},
{0.145476,0.00141863,0.129425,0.247135,0.00175521,0.0319651,0.0130426,0.00585254,0.0818154,0.0123593,0.00441789,0.040042,0.0151136,0.0906469,0.0420979,0.0656688,0.0511289,0.0169401,0.000773144,0.00292584},
{0.0169799,0.00220872,0.104366,0.037638,0.000678059,0.540447,0.0181137,0.000745968,0.0508618,0.0021147,0.000947377,0.131805,0.00387544,0.026935,0.0242347,0.029703,0.00458362,0.00158917,0.000603189,0.00157041},
{0.0402646,0.00393887,0.0293968,0.160352,0.00411534,0.012926,0.0617968,0.00984913,0.14488,0.0354919,0.0124819,0.0323829,0.00433742,0.249701,0.115202,0.0243247,0.0305645,0.0156027,0.00266766,0.00972278},
{0.225691,0.00433901,0.0637125,0.115483,0.00274401,0.0867335,0.0131067,0.00850862,0.0660469,0.0193314,0.00648772,0.0244309,0.0611149,0.0578159,0.0523417,0.107088,0.0507677,0.0299883,0.00136959,0.00289829},
{0.00331642,0.00518907,0.000826647,0.000513413,0.380176,0.00104328,0.0278451,0.00418953,0.000714992,0.0111212,0.00236218,0.00308302,0.000836508,0.00110249,0.00153108,0.00358767,0.00236089,0.00607663,0.0333347,0.510789},
{0.199501,0.058459,0.00949551,0.0125535,0.0634434,0.0939787,0.0341857,0.014021,0.0118723,0.0449388,0.0246991,0.0392255,0.0145385,0.0254265,0.0236079,0.166392,0.0533159,0.0367303,0.0129803,0.0606346},
{0.0319449,0.00660403,0.0378075,0.0279488,0.0108916,0.0365541,0.208864,0.00475423,0.0868665,0.0156546,0.00432533,0.208471,0.00608996,0.0766373,0.101167,0.0577657,0.0302051,0.00605852,0.00260019,0.0387897},
{0.07768,0.0140093,0.00806517,0.0112622,0.0173585,0.0136868,0.013373,0.123964,0.0129145,0.0724671,0.0420745,0.0403217,0.0117084,0.0179996,0.0142519,0.0922724,0.23169,0.167914,0.00281536,0.0141727},
{0.118366,0.00356248,0.0495595,0.153759,0.00686657,0.0194993,0.0229373,0.0302661,0.0982304,0.0571251,0.0171727,0.0259525,0.0175153,0.120492,0.0805193,0.0486588,0.0635796,0.0553336,0.00230083,0.00830274},
{0.052856,0.000862588,0.209276,0.402419,0.00126507,0.0155838,0.0124149,0.00548648,0.0497017,0.00902565,0.00423571,0.0264744,0.00631856,0.121241,0.0193569,0.0197263,0.0235464,0.0159764,0.000838161,0.00339487},
{0.0344366,0.0020986,0.113901,0.054178,0.00315576,0.336164,0.0461777,0.000346342,0.0667553,0.00483557,0.00197045,0.163672,0.00403698,0.0605414,0.0426222,0.0481173,0.00891481,0.00201106,0.00065101,0.00541456},
{0.115309,0.00149624,0.175552,0.174941,0.00157002,0.0394181,0.0132402,0.0056913,0.0433118,0.010141,0.00303321,0.0458477,0.16658,0.0366731,0.0151279,0.0871536,0.0468261,0.0141932,0.00075157,0.00314327},
{0.386515,0.0218928,0.00223668,0.00313878,0.00392814,0.369435,0.00146729,0.00853761,0.00188405,0.0127257,0.00805817,0.00304205,0.0158688,0.00214647,0.00375793,0.0808877,0.0305196,0.0410635,0.000992288,0.00190203},
{0.0146571,0.00243179,0.000521057,0.00148744,0.00916375,0.00209533,0.00101819,0.19139,0.00228984,0.44328,0.221743,0.00129983,0.000768515,0.00493628,0.00288413,0.00272515,0.0170997,0.0778557,0.000877938,0.0014756},
};

const double C60StatWeight[60] = {0.0169699,0.0211683,0.0276589,0.0065676,0.0141221,0.00687748,0.014691,0.00672258,0.00183967,0.0102547,0.0230896,0.0057941,0.0125395,0.0204526,0.00706296,0.0117983,0.00683347,0.0433776,0.0318279,0.0222546,0.0102265,0.0150546,0.013416,0.0148552,0.0239112,0.0128776,0.0222319,0.0247445,0.0214275,0.0115002,0.00760174,0.0130259,0.0093702,0.0467194,0.044194,0.0322263,0.0403,0.0150234,0.010459,0.0214742,0.0154958,0.010179,0.022798,0.0123205,0.00667776,0.000415008,0.0344385,0.0113663,0.0127143,0.0124324,0.0262124,0.0064995,0.0103203,0.0142464,0.02156,0.0199151,0.00389642,0.0113449,0.0128596,0.0117657};


const double CG6StatFix[6][20] = {{0.0868109,0.00800647,0.0603889,0.0746749,0.0288533,0.0449645,0.0194255,0.0573022,0.0593364,0.0689465,0.0191267,0.0334719,0.15729,0.0352166,0.0442872,0.0554703,0.0483941,0.0688085,0.00784919,0.0213756},
{0.035671,0.00705817,0.119144,0.0415939,0.0919541,0.111489,0.0117221,0.0133989,0.033734,0.0264084,0.0084815,0.0431066,0.0438924,0.0219144,0.0370221,0.108675,0.12947,0.0186288,0.0155589,0.081077},
{0.222272,0.0458389,0.0070763,0.00673509,0.0502587,0.119971,0.0106867,0.0387956,0.00726543,0.0887825,0.0221199,0.0123076,0.0192136,0.00778413,0.0100978,0.0359658,0.0690567,0.128338,0.0429625,0.0544713},
{0.149797,0.0266512,0.0102004,0.00672546,0.144465,0.113466,0.00843539,0.0469401,0.0068417,0.186584,0.0297543,0.0149152,0.0138865,0.00794368,0.00967841,0.136586,0.0254226,0.0331411,0.0197253,0.00884092},
{0.0323714,0.00739108,0.0851093,0.0413837,0.0168777,0.198619,0.0313881,0.101708,0.0658413,0.042948,0.0101795,0.0832934,0.0150223,0.0374595,0.0506401,0.0385885,0.0218451,0.0983149,0.00639853,0.0146208},
{0.0863355,0.011268,0.0548328,0.0784968,0.0383706,0.0471228,0.0260948,0.0590604,0.0661469,0.108287,0.023513,0.040988,0.0172652,0.0472062,0.0610844,0.0580417,0.0555067,0.0680348,0.0140855,0.0382588}};

const double CG6RR[190] = {1.79343,0.518028,1.24485,0.163123,1.91633,0.373638,0.130026,0.619796,0.285445,0.869732,0.333299,1.3423,1.06875,0.483594,4.759,2.12005,2.10207,0.0973215,0.156754,0.0801425,0.00415815,0.821029,0.482439,
0.770846,0.328916,0.0178637,0.546877,0.852964,0.676909,0.108203,0.115518,0.869468,3.10049,1.22138,1.73357,0.401602,1.02872,5.15786,0.0165837,0.857585,0.854771,0.0116738,0.257285,0.0153956,0.0287817,4.69939,0.388532,0.501966,
0.110999,1.20927,0.407981,0.0429497,0.0307864,0.114746,0.0164572,0.44012,0.368033,0.0434091,1.63234,0.0650548,0.163329,0.520356,0.453844,3.81647,0.315032,0.628084,0.614904,0.273699,0.0591765,0.090583,0.0543423,0.810122,1.07332,0.025083,2.18883,1.64161,0.0929775,0.0839157,0.0382749,0.0548988,0.269274,0.125388,0.593598,1.95222,8.12528,0.297476,0.00674531,0.322124,0.0354196,0.119438,1.26949,0.229701,0.308118,0.444644,1.7133,0.13577,
0.0612591,0.144322,0.0410265,0.112748,0.64945,0.394001,0.434056,4.34966,0.562568,4.3934,2.36195,0.991417,0.564987,0.126517,0.568354,5.337,0.160575,4.23836,4.73709,0.175825,0.0749186,0.0767769,0.126611,0.0690827,1.13532,10.9867,0.0851102,0.222506,0.147223,0.713218,2.13172,0.424831,3.21719,6.0063,0.800118,1.24324,0.194481,0.0411381,0.106429,6.46964,0.0757873,0.297741,0.621421,0.331532,0.164241,0.289096,1.76981,0.456506,0.288684,
0.398366,0.108237,1.6309,0.493036,0.328527,1.96986,2.01585,0.549499,0.43044,0.179979,1.62478,0.702198,4.35994,2.11874,0.0876909,0.0411825,0.560709,0.76018,0.399576,1.45184,0.612891,0.291633,0.0832659,0.0852953,2.67068,1.24408,1.08044,0.228432,0.188726,0.204612,0.914301,0.616416,0.178864,0.489291,0.250071,6.43242,0.106226,0.194668,0.319288,2.17205,0.0984022,0.167699,0.126945,0.213005,2.36832};

const double CG10RR[190] = {1.9989,0.496019,1.17406,0.200338,1.91595,0.389579,0.12205,0.595661,0.296248,0.912672,0.334839,1.38922,1.02478,0.471406,4.90288,2.25694,2.14441,0.115176,0.175123,0.0825082,0.00447673,0.952749,0.57962,0.62749,0.313543,0.0175657,0.516418,0.867975,0.641193,0.107063,0.116861,0.799796,3.26168,1.20469,1.57855,0.408419,0.955083,5.14903,0.0141958,0.886258,0.938485,0.0125341,0.264063,0.0150163,0.0266874,4.62005,0.391436,0.505101,0.116131,1.18308,0.433875,0.046222,0.0263185,0.124004,0.016392,0.43124,0.400306,0.0439626,1.61906,0.0721008,0.169109,0.532874,0.453929,3.77756,0.322427,0.617182,0.588234,0.266397,0.0629914,0.0961023,0.0458256,0.764678,1.06424,0.0238098,2.30736,1.64623,0.0733036,0.0930913,0.0371202,0.0538623,0.320516,0.136321,0.588568,1.9018,8.33397,0.330324,0.00755733,0.311052,0.0341079,0.118299,1.33386,0.228641,0.301955,0.447802,1.73284,0.153184,0.0857462,0.15169,0.0344445,0.120617,0.668318,0.365433,0.42956,4.52837,0.552732,4.59103,2.31266,0.985477,0.559329,0.122882,0.535858,4.69918,0.170744,4.05791,4.60592,0.201997,0.0638576,0.0798613,0.132034,0.0660288,1.09589,11.3024,0.0887821,0.219521,0.155815,0.733186,2.19898,0.419989,3.22546,6.05979,0.786273,1.22185,0.195856,0.0429686,0.102586,6.38912,0.0714957,0.312052,0.662812,0.335003,0.177929,0.309189,1.75107,0.494166,0.285974,0.366838,0.117709,1.6606,0.499453,0.338447,2.0247,1.99785,0.585877,0.439654,0.178547,1.65969,0.723502,4.25191,2.14972,0.0966373,0.036169,0.520729,0.755348,0.39234,1.47444,0.651889,0.25795,0.0814525,0.0905222,2.71118,1.21046,1.06516,0.229105,0.197651,0.204639,0.902502,0.617187,0.176016,0.500374,0.241584,6.43768,0.10171,0.159432,0.336762,1.93539,0.0762845,0.187546,0.127285,0.202075,2.36563};

const double CG10StatWeight[10] = {0.0264211,0.0392776,0.0415487,0.0664512,0.136813,0.112557,0.425776,0.0919425,0.0179704,0.0412423};

const double CG10StatFix[10][20] = {{0.0252068,0.0134833,0.15989,0.0189597,0.141849,0.0596335,0.0143987,0.0306495,0.0131915,0.208147,0.0411813,0.120686,0.0162565,0.0120666,0.014527,0.0346946,0.0199551,0.0336818,0.0129628,0.00857893},
{0.221193,0.0539464,0.0092461,0.008234,0.0501839,0.0507115,0.0129623,0.0326564,0.00821583,0.0810625,0.0223878,0.0147024,0.019687,0.00902954,0.0115915,0.0464507,0.0787133,0.164212,0.0487944,0.0560197},
{0.034161,0.00467607,0.0772063,0.0539357,0.08183,0.273411,0.020784,0.00712793,0.0862078,0.0165463,0.00727011,0.0705455,0.0203614,0.0420011,0.0512334,0.0383782,0.0251636,0.0115729,0.00738033,0.0702071},
{0.0656782,0.0310531,0.0202728,0.0285165,0.0285403,0.0340252,0.056325,0.0483359,0.0560228,0.186126,0.0212019,0.0391148,0.0233829,0.0328879,0.0756314,0.0445992,0.043252,0.0889073,0.0157781,0.0603489},
{0.0819335,0.0076874,0.0620977,0.075689,0.0243894,0.0456926,0.0189198,0.0702897,0.0596915,0.0578337,0.0155331,0.0334051,0.158368,0.0349808,0.0445188,0.0555695,0.0459228,0.0803412,0.00716953,0.0199669},
{0.160763,0.0068497,0.0582418,0.0896175,0.0273144,0.0394939,0.0196731,0.0495248,0.0710836,0.112725,0.0233476,0.0361018,0.0161564,0.0500832,0.060763,0.0524305,0.0427148,0.0526455,0.00752987,0.0229408},
{0.0727598,0.0110847,0.0609105,0.0710826,0.0421185,0.0601195,0.0248432,0.0652072,0.062239,0.0964712,0.0244816,0.0444254,0.0191939,0.0466248,0.0583233,0.0680507,0.0517497,0.0671871,0.0147925,0.0383345},
{0.0612726,0.0107711,0.0400844,0.107046,0.0407877,0.0277614,0.0243198,0.054593,0.0725232,0.0754894,0.0197225,0.0342619,0.0161999,0.047839,0.0596739,0.0476598,0.105136,0.100883,0.0131691,0.0408059},
{0.0280749,0.00687487,0.106104,0.0316359,0.0845976,0.0778142,0.0122481,0.0100383,0.0187139,0.0354596,0.00682296,0.0390195,0.0271078,0.0141056,0.0207825,0.140239,0.14748,0.0128632,0.147302,0.0327164},
{0.186614,0.0179691,0.00614378,0.00546742,0.0418463,0.256747,0.0048841,0.122381,0.00494276,0.163117,0.0196015,0.00882744,0.01352,0.00481516,0.00749694,0.0551498,0.019985,0.0360509,0.00917287,0.0152676}};

const double CG20RR[190] = {2.11409,0.500383,1.18828,0.16678,1.85623,0.370743,0.135651,0.601061,0.322578,0.938501,0.33749,1.3194,1.03445,0.471817,4.76924,2.26897,2.29836,0.123716,0.156915,0.0886014,0.0046542,0.886928,0.573939,0.659728,0.309439,0.0183007,0.546543,0.822671,0.651081,0.110225,0.11791,0.795566,3.08014,1.17946,1.69381,0.500569,1.01941,4.9752,0.0175796,0.904163,0.893106,0.011379,0.261827,0.0152947,0.0290069,4.65128,0.414261,0.503305,0.115212,1.24738,0.431835,0.0433887,0.0226006,0.123777,0.016545,0.43303,0.393747,0.0457583,1.59707,0.0713349,0.175671,0.525996,0.475326,3.70677,0.314266,0.628983,0.610858,0.283448,0.0624085,0.0923724,0.0457069,0.785836,1.04241,0.0255649,2.37741,1.75607,0.09122,0.0718723,0.0390928,0.0566299,0.283106,0.126955,0.59037,1.99507,8.57132,0.330512,0.00775584,0.331765,0.0307378,0.104977,1.36816,0.231568,0.30972,0.469768,1.80457,0.15155,0.0812712,0.166381,0.0380258,0.118539,0.690875,0.408378,0.420641,4.40687,0.578731,4.59631,2.43259,1.00088,0.559239,0.12129,0.511614,4.78174,0.178375,4.16054,4.71471,0.180345,0.0776842,0.0833785,0.137604,0.0689084,1.13229,10.7587,0.0693096,0.218724,0.153268,0.752626,2.17377,0.441954,3.15268,5.96654,0.804642,1.23906,0.202045,0.0437994,0.106916,6.59721,0.0772099,0.298307,0.660022,0.331023,0.172065,0.289563,1.74889,0.5428,0.306255,0.355736,0.113207,1.66886,0.505804,0.32401,1.99745,2.10504,0.632467,0.445907,0.185104,1.64517,0.726959,4.34959,2.13247,0.0917754,0.0316863,0.538816,0.781537,0.410353,1.48924,0.653165,0.296438,0.0745747,0.0681223,2.6391,1.23509,1.06965,0.236981,0.198519,0.20872,0.925152,0.61527,0.178253,0.493473,0.248647,6.40502,0.107439,0.184824,0.35306,2.05214,0.0904017,0.189369,0.112636,0.193995,2.31209};

const double CG20StatWeight[20] = {0.0151671,0.02541,0.0319625,0.105569,0.0215911,0.0118144,0.02245,0.0195248,0.021391,0.0345195,0.0101073,0.00979145,0.0262499,0.0122659,0.0332378,0.0226548,0.394049,0.0660986,0.0989422,0.0172038};

const double CG20StatFix[20][20] = 
{{0.210154,0.0186798,0.0073746,0.0055795,0.16095,0.0411903,0.00671303,0.127051,0.00494912,0.0891817,0.0197337,0.00880306,0.0072309,0.00532825,0.00720679,0.180542,0.0212324,0.0380361,0.0291636,0.0109005},
{0.112359,0.00455834,0.071773,0.117865,0.0662674,0.0477489,0.0142761,0.0186595,0.0696884,0.0257877,0.00895266,0.0281211,0.155296,0.0344675,0.0390126,0.0538755,0.038975,0.0265985,0.00776656,0.0579512},
{0.0581015,0.0154461,0.0156867,0.0249113,0.027386,0.0259554,0.0325833,0.0457675,0.0830162,0.231217,0.0178669,0.0277077,0.0158723,0.0332826,0.0929473,0.0318204,0.0382748,0.108346,0.0142525,0.0595583},
{0.0735003,0.00789947,0.0614761,0.0662592,0.0262771,0.0461376,0.0210374,0.0598062,0.0601797,0.0827893,0.0197491,0.0371155,0.138023,0.0382543,0.0503864,0.0596482,0.0505067,0.0731838,0.00680925,0.0209613},
{0.0341513,0.00545945,0.0587161,0.0576528,0.0759392,0.26299,0.02942,0.00834032,0.100917,0.0198976,0.00801009,0.083439,0.0175293,0.0490507,0.0564395,0.0345119,0.0229267,0.0126869,0.0076066,0.054315},
{0.0324293,0.00527821,0.139677,0.0382104,0.082371,0.0613269,0.0129788,0.0111244,0.0329456,0.0326612,0.00792921,0.0518243,0.0324477,0.0206753,0.0311703,0.0991893,0.148284,0.0144588,0.112119,0.0328986},
{0.0192211,0.00780208,0.177222,0.0190742,0.0231859,0.036156,0.0173114,0.115696,0.0243846,0.0243232,0.00585138,0.1219,0.0141108,0.016756,0.0357972,0.0362234,0.015455,0.123782,0.113212,0.0525358},
{0.224086,0.0303641,0.00947403,0.00930492,0.0289649,0.0283121,0.0396625,0.0282952,0.00898895,0.0579206,0.018332,0.0135574,0.00971764,0.00920601,0.0122303,0.0369203,0.0987605,0.206771,0.0316076,0.0975238},
{0.0899045,0.0174772,0.070371,0.0483721,0.0464997,0.0477694,0.116246,0.0684223,0.0360744,0.0750153,0.0201409,0.0349372,0.0180265,0.033979,0.0408902,0.0553535,0.0391512,0.0785439,0.020352,0.0424738},
{0.0240598,0.00502623,0.127367,0.0524954,0.0181562,0.242247,0.012108,0.111236,0.0166597,0.172263,0.0147831,0.0508176,0.0120407,0.0149786,0.0136986,0.030017,0.0141647,0.0557728,0.00266953,0.00943883},
{0.32485,0.00646891,0.00402635,0.00458941,0.0855932,0.375822,0.00270232,0.00756536,0.00410242,0.0114352,0.00370713,0.00452029,0.0057548,0.00298052,0.00534536,0.0262245,0.0109656,0.0115053,0.00851163,0.0933291},
{0.0702705,0.02854,0.0136209,0.0192119,0.0314353,0.0416602,0.051366,0.125009,0.0186227,0.0530298,0.0927465,0.181029,0.0186571,0.0248167,0.0259154,0.0675001,0.0357654,0.04452,0.0092607,0.0470229},
{0.110562,0.00645162,0.0656134,0.0776383,0.0215532,0.0386858,0.0186354,0.0692096,0.159271,0.0787419,0.0182859,0.0445761,0.016386,0.0473309,0.0346115,0.0563277,0.049741,0.0638178,0.00503061,0.0175299},
{0.0251301,0.00885211,0.008606,0.00593483,0.186348,0.069964,0.00581214,0.0180551,0.00618778,0.256966,0.0153252,0.0138183,0.0141526,0.00547276,0.0075588,0.164599,0.150958,0.0176633,0.0117892,0.00680665},
{0.154817,0.0294364,0.00807143,0.00778561,0.106906,0.216249,0.00742415,0.055067,0.00778879,0.126014,0.0295639,0.0138397,0.0112239,0.00885468,0.0112283,0.0522995,0.0495454,0.0771866,0.014838,0.011861},
{0.13953,0.0150301,0.00862795,0.0108223,0.0557605,0.0592855,0.00953924,0.061827,0.00965789,0.0669752,0.0165746,0.015208,0.269268,0.0102332,0.0121085,0.0406112,0.0384841,0.0859996,0.0223782,0.0520787},
{0.0690328,0.0109658,0.0545564,0.0865858,0.0400639,0.0437663,0.0246886,0.0628316,0.0648266,0.0905369,0.0212443,0.0414698,0.0174392,0.0487973,0.0623772,0.0594016,0.0657791,0.0798259,0.0149051,0.0409058},
{0.0723173,0.0084998,0.0586445,0.0528097,0.0359048,0.117435,0.0217619,0.0326558,0.0555816,0.140265,0.0451266,0.0471733,0.0176232,0.0460495,0.0542578,0.0786215,0.0405648,0.03291,0.0102029,0.0315952},
{0.159204,0.00698603,0.0621695,0.0878829,0.0280776,0.0442158,0.0202868,0.0526302,0.0629738,0.106621,0.0206209,0.0373618,0.0151963,0.0515852,0.0687167,0.0529257,0.0417233,0.0478719,0.00854669,0.0244031},
{0.0960794,0.1477,0.0165381,0.018663,0.0341546,0.0433707,0.0206899,0.0508377,0.035846,0.120712,0.0237936,0.0269431,0.0158141,0.026639,0.0559393,0.063585,0.0646062,0.0816382,0.0161519,0.0402985}};

const double CG30RR[190] = {2.02429,0.502699,1.18179,0.161901,1.87411,0.380328,0.139251,0.602231,0.329061,0.967546,0.339764,1.31259,1.04245,0.475411,4.79629,2.29305,2.29871,0.108337,0.153529,0.0895159,0.00474004,0.881958,0.585766,0.692944,0.314734,0.0181294,0.535886,0.811123,0.654756,0.110695,0.118298,0.801935,3.08135,1.22153,1.68674,0.448043,1.02688,4.9532,0.0178085,0.892249,0.896237,0.0116359,0.257757,0.0145007,0.0289616,4.65077,0.404454,0.49421,0.115588,1.23961,0.443964,0.0440033,0.0274052,0.130264,0.0167688,0.429775,0.396366,0.0445985,1.58523,0.0719802,0.178511,0.539138,0.468375,3.72752,0.317567,0.633631,0.618076,0.278951,0.0628244,0.0950746,0.0482704,0.803231,1.05297,0.0259739,2.35791,1.72167,0.0943343,0.0822182,0.0390299,0.0547259,0.277504,0.127103,0.59283,1.82881,7.94771,0.322459,0.00710413,0.326257,0.0354083,0.112543,1.3809,0.223634,0.309189,0.455351,1.7612,0.155441,0.0715037,0.169221,0.038345,0.112653,0.694631,0.416868,0.410448,4.32784,0.570193,4.53105,2.37513,0.991476,0.567517,0.126756,0.512118,4.64617,0.17331,4.06603,4.59082,0.18363,0.0814084,0.0809652,0.136452,0.0704865,1.12741,10.905,0.08509,0.22466,0.160745,0.773769,2.1749,0.42319,3.15421,6.00555,0.804877,1.23078,0.202154,0.0440114,0.108119,6.25843,0.0725231,0.304336,0.673489,0.346679,0.174624,0.297744,1.7546,0.51416,0.303853,0.392311,0.114355,1.68575,0.520828,0.32851,2.02536,2.06449,0.592622,0.447778,0.184504,1.64366,0.722644,4.36545,2.1658,0.0870911,0.0381145,0.56796,0.769104,0.403714,1.45794,0.64509,0.309873,0.0737725,0.0770744,2.65358,1.24231,1.0847,0.234686,0.199245,0.210494,0.925026,0.619234,0.182556,0.484173,0.242842,6.57444,0.10918,0.168973,0.336657,2.07975,0.0793159,0.187483,0.133619,0.214308,2.15871};

const double CG30StatWeight[30] = {0.0153426,0.202839,0.017766,0.0103904,0.0404552,0.0121959,0.0226171,0.0160274,0.0152089,0.00775648,0.0184207,0.0138605,0.0145574,0.0126643,0.232317,0.00403728,0.00856772,0.0130411,0.0104783,0.011828,0.0295193,0.0137968,0.0106506,0.0143698,0.0162982,0.0256894,0.0314999,0.0180015,0.0898437,0.0499597};

const double CG30StatFix[30][20] = 
{{0.0238436,0.00939411,0.196689,0.0143555,0.0186079,0.0289224,0.0106599,0.0569409,0.0102681,0.242949,0.0147166,0.131628,0.0127284,0.00893618,0.013194,0.0291075,0.013605,0.143212,0.00995295,0.0102882},
{0.0701207,0.0107927,0.0519031,0.094246,0.0461966,0.0410956,0.0264791,0.0648495,0.0650885,0.0792002,0.0214832,0.0396416,0.01815,0.0455434,0.0617445,0.0621501,0.0596016,0.0777077,0.0172749,0.0467311},
{0.148992,0.00561603,0.0618495,0.0872568,0.0194244,0.0368592,0.0170019,0.0671006,0.158025,0.0657809,0.0166397,0.0426687,0.0160003,0.0474486,0.0345084,0.054,0.0452827,0.0560791,0.00475151,0.0147156},
{0.0281006,0.00482267,0.149194,0.0341192,0.0863042,0.0577299,0.0136027,0.00894166,0.0258334,0.0276022,0.00630277,0.0552436,0.0310347,0.0187527,0.0276906,0.108898,0.124886,0.0114992,0.143021,0.0364205},
{0.0246404,0.0045903,0.0775182,0.0444655,0.00848799,0.234244,0.0298431,0.107436,0.0784911,0.0289648,0.00682426,0.0810706,0.0104729,0.0405597,0.0529869,0.0305062,0.0165244,0.10971,0.00381884,0.00884415},
{0.169261,0.0297287,0.00717978,0.00668,0.0320076,0.0388276,0.0108716,0.0340005,0.00631861,0.0655918,0.0197232,0.0109414,0.0101538,0.00748268,0.00798345,0.0446547,0.185942,0.120053,0.104101,0.0884967},
{0.132703,0.0149806,0.00872565,0.0104803,0.0585485,0.0588917,0.0104455,0.0614118,0.00952578,0.071236,0.0175125,0.0155969,0.269885,0.0105634,0.0119618,0.0422008,0.0389236,0.0814026,0.022183,0.0528223},
{0.0677832,0.0218362,0.0173241,0.027628,0.0389886,0.0362829,0.0459095,0.0624803,0.0393804,0.119478,0.027113,0.168218,0.0154806,0.0332576,0.0513253,0.050074,0.0462045,0.0688135,0.0130783,0.0493446},
{0.058675,0.0179327,0.0171988,0.0271706,0.0366981,0.0405139,0.0490998,0.0521738,0.0370582,0.125678,0.0198237,0.0268838,0.0218583,0.034542,0.157975,0.0453035,0.0673757,0.0905065,0.0248353,0.0486975},
{0.292653,0.0834196,0.0060149,0.00480738,0.026905,0.0557761,0.00502979,0.0483719,0.00576582,0.217651,0.0215878,0.0113076,0.00516664,0.00600029,0.00936005,0.0301394,0.0472809,0.0935586,0.00995492,0.0192491},
{0.0408596,0.00851546,0.0504025,0.0806719,0.0335383,0.0235132,0.0241235,0.0764785,0.0825354,0.0748974,0.0197942,0.0365758,0.0161907,0.0438009,0.0550077,0.0368806,0.181709,0.0706058,0.01331,0.0305892},
{0.0732812,0.0107759,0.0279298,0.131199,0.0228006,0.0235618,0.0194997,0.0375839,0.0836882,0.0613804,0.0189887,0.0268952,0.0188976,0.0556169,0.0522237,0.0423965,0.0876545,0.181712,0.00436143,0.0195533},
{0.0762959,0.145161,0.0183265,0.0222801,0.0385468,0.0432718,0.0241947,0.0527691,0.0486352,0.0888738,0.0263979,0.0275817,0.0173522,0.0324683,0.0595505,0.0729397,0.0534613,0.0899506,0.0209866,0.0409564},
{0.0511759,0.0168505,0.0142517,0.0232758,0.0396879,0.0295667,0.0250702,0.033687,0.118201,0.255487,0.0266875,0.0203688,0.0143691,0.0333384,0.0890804,0.0300975,0.0295948,0.0427006,0.0164999,0.090009},
{0.0755412,0.0100671,0.0673535,0.0670227,0.0346282,0.0543306,0.0239222,0.0608131,0.0653211,0.119673,0.0245306,0.0451793,0.0183254,0.0511685,0.0572793,0.0619029,0.0512327,0.0655453,0.0123266,0.0338362},
{0.0332289,0.0130111,0.130341,0.0296747,0.0158405,0.0376155,0.15029,0.166796,0.013583,0.0453253,0.126259,0.025958,0.0129995,0.0848526,0.0135556,0.0257042,0.0239954,0.0346628,0.00511641,0.0111902},
{0.278608,0.0316317,0.00775373,0.00704498,0.0627178,0.0362344,0.00747483,0.0249922,0.00794611,0.0459245,0.0188519,0.0111616,0.0123412,0.00674481,0.0105436,0.0277417,0.043644,0.281322,0.0197574,0.0575626},
{0.148562,0.0047189,0.0640532,0.125225,0.0209986,0.0385655,0.0152805,0.0357331,0.0684518,0.0450287,0.0116211,0.0281767,0.161422,0.0363573,0.0405683,0.0501007,0.0397104,0.042692,0.00572277,0.0170116},
{0.0227041,0.00789188,0.00830538,0.0053434,0.198157,0.0544427,0.00453139,0.0172929,0.00566532,0.268077,0.0150259,0.012899,0.01225,0.00488885,0.00660328,0.163018,0.155444,0.016488,0.0138775,0.00709428},
{0.0469248,0.0144938,0.0346646,0.0152263,0.0328242,0.212405,0.0143961,0.0230347,0.0109979,0.197639,0.0875311,0.0397069,0.019283,0.0129148,0.013184,0.15578,0.0219971,0.0205289,0.00983633,0.0166321},
{0.133406,0.0254556,0.0090988,0.00840602,0.11611,0.230179,0.00829594,0.0567851,0.00821503,0.122774,0.0344875,0.0145304,0.0128844,0.0100794,0.0119112,0.0542445,0.0460827,0.0703982,0.014022,0.012634},
{0.217362,0.0188233,0.00623719,0.00510468,0.160862,0.0349352,0.00543453,0.122538,0.00456405,0.08238,0.019967,0.00779851,0.006855,0.00494594,0.00659142,0.17775,0.0210708,0.038302,0.0464078,0.0120705},
{0.313745,0.00608969,0.00413218,0.00461299,0.0928298,0.37839,0.00296195,0.00586899,0.00390147,0.00969964,0.00304969,0.00451278,0.00592923,0.00283758,0.0049236,0.0262657,0.0105522,0.0102247,0.0086472,0.100825},
{0.0614721,0.00521216,0.0921074,0.0779496,0.0105491,0.0758189,0.0180772,0.013645,0.0992198,0.0206772,0.00738481,0.0381335,0.252351,0.0366857,0.0444312,0.0639273,0.0459807,0.0188897,0.00713836,0.010349},
{0.136884,0.0174437,0.0437955,0.0367344,0.0365502,0.0494152,0.103991,0.0675198,0.025972,0.0773116,0.0208846,0.0285242,0.0138143,0.0233754,0.0264957,0.0638426,0.0399232,0.0801864,0.0264636,0.0808732},
{0.193976,0.00551518,0.0809404,0.0894555,0.05611,0.0540582,0.0218126,0.0275429,0.0636513,0.0439413,0.0126334,0.0389409,0.0162152,0.0512322,0.075838,0.0531698,0.0384462,0.0259653,0.0141827,0.0363729},
{0.0739945,0.00635535,0.0607442,0.071211,0.0797361,0.18157,0.0194834,0.0140884,0.0656168,0.0249284,0.0107016,0.0422177,0.030321,0.0435832,0.058728,0.0728786,0.0483461,0.0194137,0.00962885,0.0664529},
{0.0231371,0.00540847,0.141637,0.12843,0.0216004,0.149261,0.00760253,0.142178,0.0133023,0.201419,0.0156495,0.0208914,0.0162048,0.0133935,0.0107266,0.0216332,0.0133006,0.0377939,0.00385299,0.0125778},
{0.0766444,0.00814295,0.0580542,0.0687741,0.0295623,0.0406286,0.0218404,0.058251,0.0563593,0.08622,0.0211464,0.0357257,0.133282,0.0382338,0.0505312,0.0586862,0.051006,0.0732785,0.00788015,0.0257529},
{0.144948,0.0081771,0.0450677,0.0971048,0.0247819,0.0368458,0.0193982,0.0510416,0.0592379,0.133703,0.0241775,0.0331331,0.0138009,0.0525104,0.0727899,0.0500609,0.0449086,0.0579063,0.00766175,0.0227447}};

const double CG40RR[190] = {2.12217,0.503926,1.18578,0.173662,1.99745,0.375597,0.130289,0.600891,0.315447,0.974427,0.341716,1.31546,1.04178,0.475832,4.82756,2.30019,2.31143,0.115171,0.164569,0.0876139,0.00467958,0.886295,0.598855,0.672875,0.31569,0.0179567,0.554142,0.8082,0.655129,0.110071,0.116566,0.792975,3.0848,1.19461,1.71149,0.44818,1.01622,4.93204,0.0149826,0.912075,0.891024,0.0131022,0.25893,0.0158096,0.0286303,4.64195,0.408423,0.503391,0.114408,1.23433,0.444318,0.0475804,0.0272533,0.115851,0.016362,0.449137,0.395337,0.0484068,1.59446,0.0721862,0.171026,0.53977,0.469415,3.73883,0.317616,0.635304,0.614524,0.286041,0.0618362,0.0913531,0.0568102,0.771869,1.0566,0.0252126,2.35477,1.70651,0.0830369,0.0756267,0.0384776,0.0555207,0.29728,0.132488,0.589639,1.74254,8.29733,0.329984,0.00590078,0.3356,0.0307506,0.107171,1.40822,0.232511,0.313423,0.468012,1.81969,0.157843,0.0657661,0.183241,0.0456408,0.12067,0.699506,0.407497,0.443735,4.33737,0.570003,4.49524,2.40139,1.00828,0.571722,0.125984,0.528472,4.77292,0.180204,4.18196,4.68253,0.192733,0.0825132,0.0845518,0.137344,0.0682317,1.14993,11.1301,0.0905509,0.225436,0.159829,0.793921,2.17981,0.436602,3.18377,6.04153,0.803298,1.2426,0.201151,0.0425982,0.105731,6.32891,0.076606,0.308744,0.669219,0.343771,0.189489,0.316026,1.77375,0.499831,0.294454,0.394178,0.117567,1.71343,0.530052,0.334902,2.08675,2.08704,0.602524,0.443115,0.186093,1.65849,0.721742,4.35485,2.17936,0.0926852,0.0376483,0.543974,0.775039,0.412021,1.46901,0.644584,0.309975,0.0739819,0.0683957,2.67813,1.24897,1.09527,0.238322,0.19603,0.206394,0.92969,0.62089,0.179787,0.492174,0.247292,6.59788,0.108612,0.150532,0.334381,2.10875,0.0670639,0.17925,0.140445,0.213476,2.26885};

const double CG40StatWeight[40] = {0.00810409,0.0133755,0.050712,0.0135746,0.0226221,0.0487758,0.016721,0.00286897,0.0130862,0.0102669,0.00652415,0.00721928,0.0640191,0.0139998,0.0108572,0.00539617,0.0148846,0.00516825,0.00685468,0.00494693,0.024571,0.0114408,0.00310448,0.00746654,0.0851399,0.34477,0.0100322,0.0084026,0.00790613,0.0323001,0.00842155,0.011786,0.0096757,0.0135412,0.0113355,0.00788202,0.0172462,0.00618516,0.0242006,0.0146149};

const double CG40StatFix[40][20] = 
{{0.0520413,0.019213,0.0140244,0.0218542,0.0342169,0.0414858,0.0525237,0.0530213,0.0300167,0.124981,0.0193445,0.0251652,0.0251958,0.0317007,0.200276,0.0424081,0.0511524,0.0821682,0.0272928,0.0519182},
{0.0680126,0.00557639,0.0937984,0.0632506,0.0146351,0.106585,0.0194282,0.0146437,0.0912162,0.0227247,0.00796778,0.0393708,0.21802,0.0371457,0.0479645,0.0672561,0.0426671,0.0196734,0.00674999,0.0133139},
{0.058243,0.0105248,0.038958,0.118371,0.0406724,0.0289017,0.0280178,0.058376,0.0616738,0.0731638,0.0191525,0.0334186,0.0146782,0.0471239,0.0707491,0.0499138,0.0964801,0.0937537,0.0140144,0.0438138},
{0.0723789,0.0186378,0.0172516,0.0289023,0.0333063,0.0319859,0.0479012,0.0572833,0.0423952,0.134729,0.0231263,0.167299,0.0137589,0.034264,0.0554925,0.0440359,0.0479379,0.0691693,0.0108648,0.0492797},
{0.140483,0.0314743,0.00969368,0.00916005,0.125807,0.191936,0.00966864,0.0512554,0.00915757,0.112099,0.0348002,0.0166883,0.0155382,0.0107055,0.0132298,0.0582549,0.0547051,0.0748739,0.0180491,0.0124203},
{0.148668,0.00793134,0.0467575,0.0985171,0.0261082,0.0335013,0.0200238,0.0494704,0.0582048,0.133752,0.0238382,0.0341019,0.0137129,0.0512929,0.0739847,0.0500566,0.0435871,0.0561538,0.0076296,0.0227081},
{0.166767,0.00522377,0.0612284,0.0922746,0.0207358,0.0345568,0.0172397,0.0680279,0.144261,0.0669428,0.0164691,0.0414591,0.0144323,0.0487753,0.0361914,0.0524818,0.0437221,0.0498679,0.00448499,0.0148591},
{0.0470715,0.0165211,0.0189874,0.0241318,0.043848,0.0738661,0.071161,0.0262812,0.0178681,0.182692,0.0539728,0.0216823,0.0201313,0.239735,0.0191632,0.0306678,0.0288945,0.0336244,0.0105801,0.0191197},
{0.220284,0.0183242,0.00660464,0.00510928,0.167718,0.0318231,0.00571518,0.12329,0.00452765,0.0910783,0.0186773,0.00752385,0.00690252,0.00482804,0.00668765,0.17773,0.0208588,0.0370894,0.0349876,0.0102415},
{0.226067,0.0259802,0.0103911,0.00999094,0.0232214,0.0370227,0.0703817,0.0508036,0.0116342,0.147737,0.020527,0.0172374,0.00696564,0.0127519,0.0136119,0.0347791,0.0352966,0.112242,0.0182052,0.115154},
{0.0343574,0.00907948,0.00855887,0.00889099,0.0176209,0.0332865,0.00830466,0.0398369,0.124321,0.305517,0.0139473,0.00977124,0.0152891,0.0140234,0.115465,0.015087,0.0169085,0.188117,0.0115021,0.0101155},
{0.0264484,0.00486277,0.15666,0.0346032,0.0409039,0.038212,0.0128485,0.0102461,0.0256319,0.0270426,0.0049647,0.0571234,0.0392009,0.018226,0.0273179,0.115048,0.129555,0.0119237,0.137712,0.0814689},
{0.0738331,0.00905205,0.0530077,0.0647854,0.0304129,0.122991,0.0228141,0.0643244,0.0560472,0.102611,0.0264318,0.0455084,0.0192593,0.0469301,0.060027,0.0663768,0.0458085,0.056486,0.00882487,0.024468},
{0.0658338,0.0174934,0.0199088,0.0340323,0.034199,0.0259993,0.0296058,0.035354,0.100432,0.242248,0.0255387,0.0260436,0.0128369,0.0467962,0.0677223,0.0394334,0.0408271,0.0489791,0.0148132,0.0719028},
{0.0165922,0.00705538,0.197096,0.012373,0.186794,0.0207849,0.0060436,0.0261296,0.00631277,0.257698,0.0212436,0.139407,0.009027,0.00612063,0.00724186,0.0203711,0.014333,0.0268329,0.0120295,0.0065135},
{0.0332203,0.0128607,0.0218344,0.0249172,0.0391467,0.0175426,0.0205652,0.111366,0.0727401,0.0567816,0.0229279,0.0265284,0.0201994,0.0261454,0.0644574,0.025756,0.274535,0.0619797,0.011371,0.055125},
{0.0873191,0.152638,0.0183687,0.0216866,0.0352961,0.0404637,0.0237228,0.0523163,0.0465181,0.103184,0.0251534,0.0267624,0.0155297,0.0307268,0.0605158,0.0671218,0.0505957,0.0830118,0.0174748,0.041594},
{0.0579367,0.016158,0.0131742,0.00817218,0.0209621,0.228723,0.00670795,0.173132,0.00823042,0.0534562,0.108586,0.0185494,0.010213,0.00866164,0.008356,0.18024,0.0220819,0.0322394,0.00481861,0.0196016},
{0.0573218,0.00316546,0.0782036,0.1505,0.0868572,0.0315086,0.0084578,0.00766338,0.0502086,0.0118368,0.00494787,0.0224994,0.251304,0.0275624,0.029062,0.0440198,0.0354864,0.0137829,0.00521462,0.0803964},
{0.0369104,0.00396457,0.066419,0.101614,0.0741133,0.0309353,0.0178388,0.0149476,0.137483,0.0225574,0.0077247,0.0571479,0.0186556,0.0403034,0.0362035,0.0543619,0.138874,0.0206977,0.0651194,0.0541282},
{0.18721,0.00544478,0.0835608,0.0860459,0.0631235,0.0580323,0.0208896,0.0243743,0.061062,0.0394324,0.0118693,0.0382003,0.0160411,0.0508638,0.0775492,0.0543519,0.0382026,0.0243818,0.0143598,0.045005},
{0.176882,0.0407242,0.00761194,0.0066424,0.0368311,0.0355942,0.00857384,0.0339682,0.0065179,0.111044,0.019828,0.0105874,0.0092781,0.00760756,0.00931308,0.0419536,0.190199,0.106402,0.092098,0.0483441},
{0.0342941,0.0343543,0.0221947,0.0145951,0.0520582,0.0323259,0.016394,0.0472928,0.0109018,0.051239,0.122872,0.147549,0.00832396,0.0140727,0.0151686,0.179251,0.023863,0.142836,0.0112448,0.01917},
{0.0331793,0.00385996,0.0553737,0.0695779,0.0848483,0.230838,0.0195585,0.0066366,0.157459,0.0185177,0.00847826,0.0677394,0.0121745,0.0491965,0.0436706,0.0337946,0.0211961,0.0108752,0.00522876,0.0677966},
{0.0822648,0.00763002,0.059615,0.0763254,0.0287686,0.0385771,0.021334,0.0574897,0.0617276,0.0812468,0.0200342,0.0355888,0.128121,0.0391938,0.0513802,0.0577481,0.0499316,0.0718456,0.00681306,0.0243641},
{0.0759885,0.0105536,0.0627376,0.0781936,0.0401886,0.0442698,0.024735,0.0610305,0.0657117,0.102135,0.0226869,0.0425952,0.0189169,0.0485978,0.0584298,0.0619766,0.055191,0.0693325,0.0157993,0.0409302},
{0.0599278,0.00752376,0.0523601,0.0721384,0.0499465,0.0537746,0.0219729,0.0746754,0.0840427,0.0486587,0.0150169,0.0375573,0.0155772,0.0362304,0.0509637,0.14337,0.0373884,0.0662811,0.00943224,0.0631621},
{0.0413739,0.00570945,0.145295,0.0389812,0.0845615,0.243918,0.0105423,0.0063084,0.0243063,0.0139945,0.00565239,0.036724,0.0185392,0.0232147,0.0278591,0.0710839,0.100157,0.0120155,0.00722241,0.0825422},
{0.22544,0.013757,0.00576032,0.00784642,0.064292,0.0465999,0.00466478,0.0670063,0.00633007,0.0356337,0.0118014,0.00773081,0.279709,0.00599314,0.00719816,0.0318779,0.0259238,0.0705094,0.0114836,0.0704429},
{0.0229525,0.00464095,0.0843456,0.0452714,0.00705029,0.2453,0.032633,0.108374,0.061257,0.0251018,0.00599257,0.0842759,0.0109845,0.0396854,0.0525849,0.0306103,0.0156954,0.110813,0.00367872,0.00875273},
{0.280586,0.0345691,0.00833595,0.0066832,0.0610699,0.0350307,0.00653088,0.0226692,0.00720209,0.040795,0.0168215,0.0116308,0.0123942,0.00632732,0.0100938,0.0286314,0.0499981,0.279137,0.0236133,0.0578807},
{0.0877308,0.0153171,0.0780932,0.0481247,0.0652893,0.0411332,0.140441,0.0605391,0.0356989,0.0652407,0.0216385,0.0341753,0.0178322,0.030081,0.0327158,0.0537854,0.0403866,0.0753546,0.0209631,0.0354597},
{0.0832862,0.0126801,0.0290739,0.107836,0.0231325,0.0226361,0.0195018,0.0330063,0.103664,0.0510426,0.0168192,0.0278368,0.0176713,0.0537752,0.0447064,0.0440934,0.0792841,0.207645,0.00395806,0.018351},
{0.0382293,0.0143116,0.12609,0.0330414,0.0697488,0.0285659,0.0216971,0.0855716,0.0683079,0.0423922,0.0138145,0.0819752,0.0223792,0.0326013,0.0727054,0.052112,0.0315424,0.0997644,0.0133103,0.0518401},
{0.0262281,0.0136862,0.140199,0.149588,0.0387319,0.0980033,0.0112242,0.0237615,0.0149693,0.216283,0.0855304,0.0190134,0.0156198,0.0146818,0.0136215,0.0248677,0.0171811,0.0229771,0.0133636,0.040469},
{0.172235,0.00437536,0.0648551,0.117325,0.0203067,0.0337935,0.014481,0.0325745,0.0754791,0.0436332,0.0107095,0.0274357,0.158166,0.0372617,0.0387508,0.0478547,0.0384939,0.0413994,0.00501881,0.0158512},
{0.229634,0.0102478,0.00347106,0.00380115,0.0187557,0.293449,0.00258105,0.139481,0.00293053,0.18508,0.0145325,0.00467376,0.00448198,0.00268628,0.00443753,0.029396,0.0105331,0.0273595,0.00301365,0.00945445},
{0.0204045,0.00631159,0.00800439,0.00544478,0.141705,0.0403678,0.00532951,0.00558855,0.00478439,0.0286913,0.00474876,0.0115764,0.00960929,0.00441973,0.00901434,0.183109,0.182564,0.00751362,0.305985,0.0148284},
{0.0725904,0.0144415,0.0149636,0.0168662,0.0465565,0.0500271,0.0196174,0.0560454,0.0156415,0.0977582,0.0198294,0.0280775,0.241042,0.0201243,0.0197088,0.0574855,0.0576544,0.084227,0.0239677,0.0433756},
{0.0144177,0.00444978,0.0160635,0.00447944,0.0151234,0.528928,0.00178647,0.0878109,0.00202152,0.17028,0.0134353,0.00808987,0.00685251,0.00254821,0.00284449,0.0174064,0.00643841,0.0889414,0.00349958,0.00458304}};

const double CG50RR[190] = {2.03049,0.504754,1.18106,0.176222,1.89564,0.378192,0.141708,0.603419,0.330212,0.950535,0.341057,1.31696,1.03739,0.473729,4.77952,2.29473,2.32627,0.104394,0.152707,0.088683,0.00469828,0.910732,0.58788,0.697425,0.314919,0.0181307,0.542669,0.831487,0.657149,0.109296,0.117216,0.799117,3.07713,1.18014,1.69238,0.457301,1.05274,5.01875,0.0148698,0.908473,0.89823,0.012098,0.262369,0.014226,0.0282929,4.71787,0.417104,0.508355,0.115388,1.24846,0.44745,0.0459333,0.0278088,0.126761,0.0157006,0.44865,0.393497,0.047503,1.57927,0.070686,0.179361,0.545735,0.471375,3.73648,0.316227,0.63433,0.617146,0.282349,0.0650709,0.0966627,0.0543156,0.77061,1.06125,0.0258938,2.36068,1.72266,0.0851201,0.0748267,0.0391012,0.0559806,0.309739,0.136166,0.589837,1.75677,8.20694,0.329322,0.00634532,0.340171,0.0335622,0.10874,1.40896,0.231758,0.314721,0.466817,1.77435,0.156214,0.0673098,0.183702,0.0391402,0.117672,0.694314,0.419971,0.447983,4.32247,0.566547,4.5096,2.37735,0.991946,0.567634,0.127338,0.521382,4.70861,0.177473,4.14251,4.66531,0.176877,0.0835911,0.0837714,0.137849,0.0689384,1.14208,11.022,0.0858314,0.221245,0.157262,0.775773,2.19191,0.43297,3.17237,6.02359,0.807458,1.24237,0.199792,0.0448579,0.110005,6.37191,0.0725243,0.306182,0.660228,0.343389,0.186884,0.314517,1.76299,0.522463,0.311338,0.382589,0.117454,1.68561,0.526442,0.333687,2.05024,2.07433,0.615461,0.451775,0.186243,1.66908,0.729296,4.38486,2.15927,0.0918355,0.0381091,0.57255,0.778316,0.409123,1.47032,0.63879,0.313027,0.0777718,0.0704034,2.66525,1.24825,1.08518,0.235889,0.20469,0.21272,0.927587,0.621994,0.180421,0.499435,0.248503,6.58655,0.109045,0.141456,0.327775,2.09929,0.0705747,0.184347,0.136495,0.214855,2.18665};

const double CG50StatWeight[50] = {0.00900257,0.0116695,0.00671684,0.00671106,0.0213038,0.0239525,0.00906347,0.0224708,0.00935927,0.000832967,0.00922964,0.00368689,0.121198,0.00459856,0.0454193,0.0674107,0.00613985,0.0290999,0.00253986,0.0343031,0.00632091,0.206561,0.00770799,0.0863944,0.00426762,0.0141796,0.00936515,0.00424548,0.00568493,0.0186798,0.00736598,0.00537841,0.00913881,0.00681357,0.00279223,0.011548,0.0142288,0.00859468,0.0247643,0.00920131,0.018344,0.011497,0.0120652,0.0054294,0.00358691,0.00957676,0.002896,0.0113598,0.015397,0.00190672};

const double CG50StatFix[50][20] = 
{{0.0214347,0.00806279,0.221178,0.0122703,0.167271,0.0131425,0.00762764,0.07403,0.00774683,0.133953,0.0222348,0.140943,0.0123039,0.00736134,0.012993,0.0217754,0.0132751,0.0799658,0.0153556,0.0070755},
{0.225067,0.0194698,0.00598473,0.00498406,0.0722669,0.0289724,0.00624775,0.083929,0.00484461,0.11471,0.0191127,0.00651688,0.00653398,0.00507116,0.00709827,0.184685,0.0207273,0.0368793,0.120812,0.0260865},
{0.0486098,0.0202791,0.0121077,0.0195674,0.0335347,0.0394645,0.056484,0.0537679,0.0283202,0.130187,0.0188516,0.0244565,0.0262858,0.0295471,0.211545,0.0388899,0.0529205,0.0836983,0.0201461,0.0513365},
{0.229698,0.0134534,0.00541598,0.007715,0.0639091,0.0454587,0.00498395,0.063225,0.00632309,0.0422893,0.0121949,0.00775737,0.294116,0.00589434,0.0074369,0.0319538,0.0247838,0.0642133,0.0101692,0.059009},
{0.0711323,0.00558019,0.0822729,0.058697,0.0750759,0.117432,0.0168943,0.0103449,0.0592318,0.0176644,0.00731667,0.037574,0.159831,0.0332996,0.0489148,0.0689579,0.0398562,0.0160687,0.00751868,0.0663369},
{0.0828167,0.00792349,0.065875,0.0734199,0.0256692,0.0399903,0.0218061,0.0706354,0.13911,0.103563,0.0199927,0.0443485,0.0177537,0.0461386,0.0376396,0.0558978,0.0509642,0.0675227,0.00646651,0.0224666},
{0.326661,0.00571154,0.00377205,0.00454124,0.0915568,0.379742,0.00275726,0.00558186,0.00376458,0.00990687,0.00285698,0.00418164,0.00412098,0.00272639,0.00466392,0.024581,0.00980078,0.00960349,0.00760846,0.0958614},
{0.0737151,0.0151647,0.0136943,0.0163712,0.0447384,0.0454326,0.0193122,0.0548751,0.0149264,0.0998861,0.0198665,0.0262036,0.248271,0.0192554,0.0185473,0.0553653,0.0592277,0.0891102,0.0229685,0.0430685},
{0.083985,0.0142698,0.0884376,0.0459966,0.0715223,0.040113,0.145091,0.0507898,0.0350215,0.0742728,0.0238076,0.0318826,0.0173944,0.0319073,0.0311057,0.0511689,0.0414744,0.071638,0.0168443,0.0332775},
{0.0237196,0.00244913,0.0632928,0.0342324,0.0113323,0.0782817,0.0126045,0.0131899,0.328053,0.101558,0.0846367,0.108678,0.0140162,0.031974,0.0233979,0.0257528,0.017478,0.0124345,0.00500658,0.00791196},
{0.165626,0.00468655,0.0659397,0.118136,0.0213418,0.0367859,0.0148564,0.0352276,0.0655068,0.0465158,0.0116687,0.0272495,0.157591,0.0360739,0.0396488,0.0486896,0.0385449,0.043329,0.00562515,0.0169571},
{0.0553796,0.0151978,0.0178629,0.0332022,0.0281072,0.0445072,0.0446039,0.0391849,0.0210838,0.196846,0.0316852,0.0311958,0.0144065,0.217685,0.0241753,0.0407493,0.0469361,0.0606757,0.0154481,0.0210664},
{0.0723329,0.0116056,0.0445291,0.10178,0.0444334,0.0381329,0.0285681,0.0605526,0.0614937,0.0781676,0.0207096,0.0377857,0.0173276,0.0457534,0.067275,0.0586363,0.0641554,0.0839844,0.0165268,0.0462499},
{0.0131766,0.00346865,0.00191692,0.0015176,0.0692599,0.525575,0.00315431,0.104224,0.00142764,0.0286054,0.00764532,0.00353881,0.00300401,0.00165871,0.00284608,0.0126751,0.00639735,0.115323,0.00990813,0.0846783},
{0.155477,0.00758695,0.0484461,0.0980626,0.0246501,0.034977,0.019243,0.05285,0.0617497,0.127566,0.0231069,0.0338875,0.0139008,0.0511721,0.0712255,0.0500262,0.0430288,0.0549225,0.0069556,0.0211657},
{0.0775056,0.00910829,0.0547977,0.0711779,0.0302817,0.113439,0.0226071,0.0665231,0.060111,0.0977125,0.0239967,0.0448015,0.0184147,0.0489651,0.0582821,0.064203,0.0465052,0.0587052,0.00882817,0.0240348},
{0.0321994,0.0078224,0.00786936,0.00841843,0.0165229,0.0354359,0.00734545,0.0402675,0.125662,0.308329,0.0138118,0.00839584,0.0144099,0.0134804,0.117529,0.0147605,0.0157826,0.192091,0.00997537,0.00989139},
{0.0267322,0.00465574,0.0635923,0.052317,0.00684803,0.227009,0.0331728,0.105909,0.0893813,0.0242623,0.00646382,0.075582,0.0102979,0.0477259,0.0570883,0.0316721,0.0175668,0.107919,0.00333718,0.00846774},
{0.0293332,0.0104953,0.012344,0.0339955,0.0473496,0.0238437,0.15154,0.103552,0.0185362,0.0363438,0.00991661,0.210223,0.0361882,0.0228873,0.0256944,0.0252337,0.0177471,0.107199,0.0446856,0.0328921},
{0.0650593,0.00840784,0.0544778,0.0808585,0.0484164,0.0532176,0.022325,0.072853,0.0756376,0.0783673,0.0204155,0.0407797,0.017791,0.0425433,0.0531104,0.0983502,0.04653,0.0593344,0.0120422,0.0494831},
{0.284729,0.0928228,0.00625725,0.00473709,0.0235895,0.034146,0.00574851,0.0439031,0.00561415,0.235211,0.0210194,0.0114054,0.00469191,0.00609934,0.00987391,0.0311627,0.0556536,0.0971795,0.00944299,0.0167125},
{0.0786588,0.0105645,0.0665165,0.0734763,0.0380304,0.0428019,0.0240298,0.0570044,0.0627705,0.117331,0.0237181,0.0419032,0.0181785,0.0501791,0.0597012,0.0597243,0.0556595,0.0679299,0.0131181,0.0387042},
{0.283772,0.0321986,0.00822359,0.00687384,0.0610738,0.0362096,0.007898,0.0246737,0.00782655,0.0466384,0.0183951,0.0111683,0.0125139,0.00662978,0.0105295,0.0263809,0.040523,0.283759,0.0211505,0.0535622},
{0.0813537,0.00782688,0.0604028,0.0744483,0.0284738,0.0394368,0.0215789,0.0584659,0.0618185,0.0818131,0.0204753,0.0356692,0.127384,0.039371,0.0514589,0.0579755,0.0500933,0.0713014,0.00708548,0.0235677},
{0.0905891,0.0160723,0.0197082,0.0278373,0.0503143,0.0708491,0.0201832,0.0430596,0.0389752,0.0429205,0.170478,0.0461859,0.00906362,0.0609074,0.039501,0.0658186,0.0512568,0.0847728,0.0110494,0.0404581},
{0.0769554,0.141657,0.0184275,0.0231851,0.0369089,0.0427406,0.026512,0.0514645,0.0520563,0.0880965,0.0246093,0.026191,0.0168491,0.0329783,0.0634523,0.0716611,0.0534379,0.0919092,0.0187816,0.0421263},
{0.196127,0.018132,0.0188841,0.0172268,0.0245877,0.0529273,0.0834304,0.0851918,0.015334,0.0529894,0.0166967,0.020546,0.00966766,0.0154682,0.017428,0.0747181,0.0463326,0.0802842,0.0257245,0.128304},
{0.0374166,0.0347773,0.0167177,0.0102815,0.20305,0.0271997,0.0136052,0.136189,0.00904311,0.0710655,0.0243866,0.140308,0.00724407,0.00938097,0.0135349,0.159364,0.0248424,0.0419951,0.00993394,0.00966384},
{0.0196925,0.00552313,0.0074211,0.00503798,0.140972,0.0350259,0.00501918,0.00495508,0.00432932,0.0220276,0.00415403,0.0106312,0.0096784,0.00428142,0.00821978,0.186135,0.184807,0.00683733,0.321559,0.0136934},
{0.183722,0.00586283,0.0877931,0.0788525,0.0513447,0.0643219,0.0225794,0.0265987,0.0565859,0.0440291,0.012599,0.0373332,0.0158051,0.0513559,0.087361,0.0555898,0.0388636,0.0256218,0.0187846,0.034996},
{0.0441525,0.00720348,0.0629535,0.0501673,0.0794543,0.238685,0.0238527,0.00861097,0.037356,0.0218036,0.00797044,0.0388446,0.0117668,0.0285973,0.0561638,0.082544,0.109003,0.0143617,0.0103229,0.0661857},
{0.135634,0.0102837,0.00635964,0.00647479,0.0258542,0.219955,0.00701142,0.183549,0.00683777,0.0572472,0.0999204,0.0120632,0.0101317,0.00765039,0.00812786,0.11796,0.0190366,0.0285582,0.00652387,0.030821},
{0.0456402,0.0106243,0.118693,0.0279058,0.0288825,0.153711,0.0146631,0.0221417,0.0132599,0.210286,0.0875839,0.0516188,0.0202022,0.013851,0.0128704,0.0973583,0.0265474,0.0208911,0.00658424,0.0166849},
{0.0274075,0.00493751,0.172308,0.0340644,0.0706715,0.0436819,0.0123553,0.00961089,0.0269368,0.0243841,0.00497664,0.0536727,0.0389453,0.0188656,0.0273661,0.109599,0.138154,0.0116536,0.111603,0.0588071},
{0.033488,0.00365432,0.0586058,0.110597,0.0625168,0.0269472,0.0208516,0.0118424,0.146999,0.0223014,0.00742032,0.0597987,0.0160796,0.0395299,0.0362181,0.0508921,0.144923,0.016928,0.0766736,0.053734},
{0.0188824,0.0032272,0.0997179,0.0277424,0.0192499,0.203953,0.012732,0.150649,0.0170407,0.203251,0.0141193,0.111996,0.011772,0.011144,0.0135608,0.0239892,0.0110263,0.0296726,0.00178323,0.0144903},
{0.0422885,0.0145775,0.109509,0.0352328,0.030695,0.0273664,0.0269274,0.0824142,0.0590594,0.0771516,0.0181064,0.0754771,0.0174769,0.0335254,0.0690189,0.0516945,0.0355469,0.0826922,0.0552124,0.0560278},
{0.182014,0.0265617,0.00674673,0.00656692,0.0307606,0.0310612,0.00967731,0.0290588,0.00603855,0.0508792,0.0160966,0.00874649,0.0093806,0.0067142,0.00785415,0.0451626,0.189614,0.136739,0.106372,0.0939546},
{0.160315,0.0333805,0.0079452,0.00801321,0.092492,0.220132,0.00793416,0.053201,0.00820483,0.121013,0.0321201,0.0136682,0.0136108,0.00959875,0.011813,0.0472728,0.0515062,0.07297,0.0175984,0.0172106},
{0.0212982,0.00948981,0.134146,0.161912,0.196386,0.02299,0.00500057,0.0209348,0.0124627,0.265965,0.02317,0.01196,0.0179048,0.0140951,0.0107164,0.0169699,0.0148025,0.0214182,0.0116234,0.00675536},
{0.0437888,0.00817691,0.048532,0.0861886,0.0393522,0.0240316,0.0257765,0.0687433,0.0730792,0.0736122,0.0203908,0.0325254,0.0146774,0.0470077,0.065822,0.0394014,0.158562,0.0790047,0.0144053,0.0369223},
{0.203517,0.00440126,0.061856,0.105891,0.0604752,0.0387545,0.0171054,0.0262665,0.127004,0.0372463,0.0108642,0.0414034,0.0144238,0.0514598,0.0402549,0.0519748,0.0394305,0.0252181,0.0068053,0.0356481},
{0.0957745,0.015448,0.0199684,0.0314944,0.0346142,0.035888,0.035955,0.0539959,0.0506955,0.131198,0.0254862,0.136477,0.0125714,0.0389344,0.0674028,0.0470554,0.0395577,0.0654315,0.00851784,0.0535343},
{0.0552712,0.00452499,0.0661908,0.127118,0.00604743,0.0346082,0.0157723,0.0144688,0.117209,0.0192821,0.0066869,0.0314965,0.303561,0.0352357,0.0388286,0.0475496,0.0470676,0.0199645,0.0026662,0.0064505},
{0.0475493,0.0308161,0.0197135,0.0143412,0.0191954,0.0445499,0.00906242,0.0420539,0.0119714,0.186021,0.0193712,0.159302,0.0186019,0.0169961,0.0121317,0.0386223,0.197241,0.096143,0.00591606,0.0104005},
{0.0800118,0.0111459,0.0269546,0.124576,0.0217392,0.0221507,0.0180006,0.0343782,0.0970388,0.052946,0.0171654,0.0263939,0.018842,0.0550463,0.0469843,0.0425218,0.0803811,0.20353,0.00316147,0.0170323},
{0.0335464,0.0168668,0.0168352,0.0131841,0.00639734,0.296219,0.00730897,0.108284,0.0108252,0.0233018,0.00539335,0.0198052,0.014667,0.0114374,0.0135504,0.259104,0.0190453,0.114025,0.00390286,0.00630033},
{0.0612013,0.0170489,0.0173954,0.0330209,0.0358336,0.0266089,0.0315536,0.033043,0.0970651,0.253891,0.0257246,0.0239564,0.0128783,0.0423793,0.0746546,0.0355754,0.0354014,0.0427127,0.0164504,0.0836045},
{0.019283,0.0049941,0.157051,0.0474776,0.012215,0.282429,0.00800079,0.0857461,0.0116645,0.165453,0.0133793,0.0362302,0.00738223,0.010992,0.00835791,0.021361,0.00967332,0.0904807,0.00265763,0.00517248},
{0.045902,0.0513681,0.00562449,0.00664994,0.0227219,0.0224915,0.00612291,0.193085,0.0293532,0.0371005,0.0203279,0.0122309,0.0336271,0.00912338,0.0197408,0.0211941,0.392613,0.0518091,0.00254233,0.0163725}};

const double CG60RR[190] = {2.03489,0.503571,1.18407,0.173192,1.96172,0.376663,0.137494,0.595438,0.323629,0.960845,0.340616,1.30521,1.03612,0.471374,4.78695,2.30705,2.29946,0.106938,0.157653,0.089765,0.00477401,0.91609,0.581128,0.711287,0.309779,0.018249,0.538243,0.803144,0.6514,0.108035,0.118351,0.803997,3.07031,1.16369,1.6805,0.481103,1.07506,4.92883,0.0159651,0.905807,0.904108,0.0118888,0.261191,0.0146409,0.0294395,4.69152,0.416232,0.501912,0.114427,1.24471,0.44951,0.0440361,0.02834,0.120028,0.0159665,0.448725,0.394181,0.04575,1.58247,0.0718464,0.178109,0.540516,0.469448,3.72282,0.317572,0.637183,0.619518,0.282378,0.0646445,0.0931121,0.0508249,0.76055,1.07561,0.0246484,2.4145,1.71174,0.0907371,0.0763687,0.0382106,0.0548064,0.288327,0.128781,0.606355,1.81115,8.06568,0.325601,0.00659651,0.334865,0.0323737,0.105796,1.39118,0.233669,0.312958,0.464601,1.78658,0.154934,0.0686481,0.180939,0.0440332,0.120041,0.680275,0.423323,0.439281,4.32439,0.571524,4.49381,2.34678,1.00455,0.562087,0.126926,0.505877,4.68671,0.177961,4.10736,4.66042,0.182431,0.0814052,0.0833742,0.135658,0.069965,1.14193,10.9873,0.0856547,0.223929,0.162752,0.782684,2.17118,0.435208,3.15675,5.95782,0.805674,1.24853,0.204202,0.0439361,0.102574,6.36731,0.0736211,0.30358,0.671529,0.350597,0.184037,0.30687,1.75981,0.533553,0.307428,0.379542,0.118073,1.6793,0.524399,0.331537,2.06361,2.06427,0.606939,0.447349,0.186916,1.64887,0.720031,4.35848,2.15987,0.088656,0.0394138,0.549157,0.779795,0.411917,1.4572,0.638582,0.308494,0.0781403,0.0726679,2.64903,1.24534,1.08695,0.235009,0.198453,0.207635,0.923843,0.621501,0.183418,0.49174,0.239294,6.60379,0.108893,0.143796,0.318042,2.06223,0.0773033,0.183957,0.144187,0.218949,2.16707};

const double CG60StatWeight[60] = {0.00418823,0.0114729,0.0063772,0.0165134,0.00225714,0.0076382,0.00280431,0.00249213,0.0555471,0.00299391,0.0100301,0.001281,0.00963893,0.00737191,0.00214803,0.00714751,0.000972136,0.00835644,0.0107334,0.00331094,0.0118114,0.0338559,0.0894205,0.018495,0.00653946,0.0835675,0.00874911,0.00948803,0.00877228,0.0799817,0.00242747,0.00765164,0.0111173,0.0132187,0.0178862,0.00273529,0.00336731,0.00664489,0.00515887,0.160683,0.0172265,0.0138157,0.00310061,0.026549,0.00619331,0.00838417,0.0095709,0.00227887,0.0105387,0.00159332,0.0124242,0.0200322,0.0290606,0.00548089,0.00634009,0.00764041,0.00347481,0.00372489,0.0505757,0.00714892};

const double CG60StatFix[60][20] = 
{{0.269208,0.0133852,0.00551378,0.00801876,0.0130826,0.0354504,0.00439604,0.0773178,0.00668839,0.0364054,0.00980678,0.00844795,0.36152,0.00585515,0.00824929,0.0288898,0.023856,0.0699362,0.00766691,0.00630536},
{0.0631815,0.00650291,0.0744605,0.0727102,0.0201025,0.0402538,0.0200523,0.0834471,0.167258,0.0988666,0.0181843,0.0506377,0.0154051,0.0444493,0.0338524,0.0544521,0.0517279,0.0639354,0.00494362,0.0155772},
{0.0270936,0.00917095,0.026528,0.0274681,0.00957712,0.215832,0.0941376,0.105756,0.0851715,0.0252395,0.00743786,0.0450366,0.0123992,0.0292645,0.100909,0.0311174,0.0187986,0.109756,0.00749667,0.0118092},
{0.0174864,0.00462859,0.15453,0.0273598,0.0138019,0.299935,0.0086635,0.092995,0.00926426,0.158207,0.0130377,0.0546306,0.00700109,0.0079154,0.00671176,0.0223475,0.00883104,0.0847468,0.00283425,0.00507229},
{0.044325,0.0383897,0.00895238,0.0108823,0.0265507,0.0203127,0.00845446,0.178427,0.0410078,0.0390441,0.0236521,0.0135168,0.0252884,0.0123787,0.0289744,0.021754,0.368792,0.0523032,0.00774349,0.0292512},
{0.022275,0.00537013,0.144903,0.174117,0.0180324,0.0356072,0.00741307,0.173052,0.0141813,0.231644,0.0132023,0.012005,0.0362102,0.0130231,0.0153426,0.0183302,0.0122575,0.0261392,0.00889906,0.0179951},
{0.0336128,0.00357236,0.0641784,0.105953,0.0614514,0.0250031,0.0198896,0.0114949,0.140465,0.0185338,0.00681154,0.0562766,0.0169475,0.039163,0.0379935,0.0490814,0.160466,0.0167806,0.0739171,0.0584089},
{0.131704,0.0121435,0.187572,0.0326118,0.097777,0.0462222,0.152133,0.0141752,0.0176715,0.0369091,0.0145438,0.0324512,0.0126299,0.0256363,0.015376,0.069536,0.0231172,0.0182284,0.0218182,0.037743},
{0.0698355,0.0128841,0.0649562,0.0635167,0.0380988,0.0405304,0.0304192,0.0632968,0.0579961,0.0870831,0.0223984,0.0422229,0.0179129,0.0462241,0.0657172,0.0630063,0.0561853,0.0734425,0.0421623,0.0421112},
{0.0259599,0.00326484,0.0408035,0.0476138,0.0779938,0.248402,0.0287015,0.0053891,0.196565,0.0217143,0.0117958,0.0913632,0.00878545,0.0398852,0.0404984,0.0279789,0.0166897,0.00810532,0.00597299,0.0525176},
{0.150585,0.0193062,0.00774527,0.00610292,0.202725,0.125824,0.00817875,0.104107,0.00599113,0.081796,0.0249836,0.0117236,0.0120894,0.00643503,0.00916794,0.124571,0.0247147,0.0466971,0.0188843,0.00837252},
{0.02889,0.00884602,0.0263933,0.0194416,0.0422734,0.0703017,0.201523,0.0219758,0.0146608,0.159773,0.0296105,0.014706,0.0151495,0.199497,0.0152288,0.0137751,0.0289735,0.022591,0.0548255,0.0115647},
{0.0687989,0.0177438,0.00885134,0.0114732,0.0321505,0.0215269,0.00896275,0.0500792,0.0115449,0.13334,0.0173972,0.0196962,0.302351,0.0140331,0.0162833,0.041125,0.0817458,0.10357,0.0165694,0.022758},
{0.0420285,0.0138118,0.0112237,0.0175288,0.0430014,0.0277317,0.0183605,0.0290414,0.122007,0.293225,0.0246236,0.0173059,0.015669,0.0282324,0.106352,0.0244282,0.0231537,0.0346338,0.0142326,0.0934097},
{0.185938,0.0037785,0.0421537,0.107954,0.00900661,0.0267173,0.0116174,0.0325447,0.144802,0.0332627,0.00838746,0.0227651,0.169326,0.0396849,0.0337178,0.0438256,0.0372375,0.0382973,0.00213739,0.0068468},
{0.044181,0.0120308,0.0168937,0.00870621,0.013564,0.317264,0.00731223,0.0234353,0.00641257,0.212986,0.0906559,0.0192972,0.00981107,0.00686188,0.00691819,0.14737,0.0197627,0.0202185,0.00343244,0.012886},
{0.0520391,0.0151349,0.017803,0.0171742,0.0423821,0.0223936,0.00959476,0.0178426,0.0134905,0.0739602,0.0107806,0.0141118,0.312148,0.0191122,0.0112515,0.244534,0.0262192,0.0195031,0.0401793,0.0203455},
{0.293168,0.032102,0.007843,0.00668659,0.0632742,0.0329806,0.00668601,0.0249365,0.00705679,0.0467033,0.0171589,0.0115563,0.0120711,0.00627001,0.00995321,0.0272663,0.0447172,0.271693,0.0222654,0.0556116},
{0.234394,0.0179478,0.00558287,0.0051766,0.0359043,0.0196722,0.00817143,0.0828278,0.00481577,0.0658268,0.0162905,0.00680613,0.00503519,0.00526593,0.00758134,0.179148,0.0228998,0.0523337,0.135443,0.0888771},
{0.0397477,0.0477247,0.0158057,0.0130391,0.0204568,0.0344081,0.015342,0.20744,0.00959503,0.0496832,0.118768,0.147035,0.00685117,0.00916103,0.0129862,0.183131,0.0194348,0.0295692,0.00661479,0.0132074},
{0.0748523,0.152455,0.0192494,0.0243648,0.0371102,0.0404827,0.0244154,0.0554831,0.0500132,0.0864577,0.0250411,0.0275388,0.016951,0.0344166,0.065798,0.071982,0.0489229,0.0876551,0.0145853,0.0422258},
{0.158454,0.0072417,0.0453608,0.110948,0.0265664,0.034459,0.0191017,0.0578125,0.0607329,0.108114,0.0214921,0.0334571,0.014276,0.0496469,0.0740877,0.0492194,0.0430165,0.0553243,0.00767637,0.0230133},
{0.0737036,0.0107048,0.0659099,0.0762202,0.0521466,0.0448097,0.0246506,0.0642724,0.06737,0.0703113,0.0203034,0.0427631,0.0206415,0.0496211,0.0605369,0.0619802,0.058,0.0803238,0.0122183,0.0435125},
{0.241863,0.0127861,0.00360938,0.00400188,0.0748791,0.287388,0.0028158,0.0850417,0.00348113,0.143747,0.0220454,0.00514186,0.00501863,0.00340018,0.00471135,0.0262598,0.0133959,0.0405148,0.00634811,0.0135504},
{0.220857,0.0129839,0.00960571,0.00955418,0.0199261,0.0222778,0.0708578,0.0392851,0.00850013,0.0448767,0.015441,0.011911,0.00772309,0.00956383,0.0114086,0.046769,0.156843,0.104764,0.0308554,0.145997},
{0.0828264,0.00762221,0.0616316,0.0751943,0.0272302,0.0394287,0.0213199,0.058172,0.0632183,0.0826887,0.020402,0.0360933,0.125322,0.0401334,0.0523371,0.0581395,0.0502251,0.069454,0.00657638,0.0219852},
{0.0338415,0.00843298,0.11607,0.0341489,0.0275395,0.0377956,0.029164,0.101423,0.0539468,0.137184,0.0231136,0.0883512,0.0183339,0.0263377,0.057497,0.0338668,0.0257155,0.0464949,0.00881888,0.0919242},
{0.0881114,0.0148359,0.0423463,0.0893261,0.0508062,0.0310756,0.114672,0.0531527,0.0440489,0.0743347,0.0201118,0.0301687,0.0150632,0.0315608,0.0440939,0.0504047,0.0541349,0.0914067,0.0147013,0.0456445},
{0.0793836,0.0103156,0.0283198,0.127475,0.0217271,0.0231537,0.0180005,0.0353472,0.0977219,0.0519685,0.0163031,0.0283306,0.0146253,0.0567351,0.0490677,0.0450954,0.0824298,0.19241,0.00365219,0.0179378},
{0.0983632,0.00972532,0.0661056,0.0696027,0.0287386,0.0407277,0.0222378,0.0504361,0.0654214,0.148146,0.0247455,0.0397314,0.0161021,0.054453,0.0616033,0.0557842,0.0510953,0.0621564,0.00836122,0.0264633},
{0.0650834,0.0491674,0.00554172,0.00637347,0.0132829,0.0944882,0.00263293,0.022646,0.00444272,0.0474758,0.0138411,0.00619315,0.0219593,0.00492929,0.00532864,0.02274,0.291088,0.302704,0.0162661,0.00381508},
{0.0430884,0.0162915,0.00877041,0.0061247,0.195831,0.0761695,0.00534918,0.019206,0.00616154,0.267433,0.0131868,0.0142298,0.0125027,0.00590078,0.00659836,0.133253,0.12726,0.0241248,0.0115176,0.00700061},
{0.100274,0.014977,0.0191308,0.0306974,0.0334728,0.03317,0.0360789,0.0582656,0.0497924,0.117755,0.0233451,0.139412,0.0120383,0.0390865,0.0689748,0.0478077,0.0401825,0.0699265,0.0101318,0.0554806},
{0.196838,0.00571003,0.0905234,0.0807009,0.0364832,0.0545659,0.0224733,0.030659,0.0559828,0.0502832,0.012933,0.0385726,0.0151897,0.0540763,0.0951888,0.0522069,0.0388601,0.0275319,0.0151217,0.0260994},
{0.129318,0.039126,0.0105281,0.010207,0.0764096,0.193713,0.0101923,0.0459801,0.0100169,0.108475,0.0297687,0.0164815,0.0139083,0.0122044,0.0137898,0.0523074,0.0817051,0.0896605,0.029432,0.026776},
{0.0458144,0.030447,0.0172454,0.0143617,0.0200373,0.0502985,0.00825587,0.0511221,0.0122962,0.146794,0.018767,0.212507,0.0188296,0.016424,0.0120186,0.0404026,0.184248,0.0822077,0.005418,0.012506},
{0.111781,0.0147157,0.0186234,0.0249818,0.0536604,0.0830702,0.0220727,0.0335792,0.0473539,0.0441903,0.1735,0.043165,0.0090145,0.0585345,0.0455082,0.0651241,0.0510379,0.0444671,0.012215,0.0434049},
{0.0246972,0.0117623,0.125367,0.114548,0.141264,0.137627,0.0097206,0.025896,0.0168464,0.182643,0.0553451,0.0272754,0.00739417,0.0183675,0.0128063,0.0240353,0.0162292,0.0249602,0.0152561,0.00795963},
{0.0309644,0.00999775,0.11331,0.0296506,0.0768836,0.0420076,0.0191258,0.0897777,0.0570099,0.0348745,0.010633,0.0687768,0.0120384,0.0230934,0.0613009,0.144067,0.026133,0.101602,0.0133134,0.0354401},
{0.072266,0.0102184,0.0517687,0.0950703,0.0411478,0.0443873,0.0242213,0.0617904,0.0658437,0.100632,0.0223594,0.0412869,0.0178779,0.046187,0.0591997,0.0645828,0.054991,0.0693761,0.0120541,0.0447394},
{0.0618413,0.00482007,0.0888378,0.0565399,0.0811942,0.118131,0.0128993,0.00834433,0.0562705,0.0142581,0.00595626,0.0382863,0.164089,0.0318144,0.0491224,0.0711567,0.042634,0.0139728,0.00557625,0.074255},
{0.122386,0.00568562,0.05925,0.0823269,0.0788444,0.13873,0.016442,0.0146469,0.0825927,0.0235032,0.00914579,0.0379757,0.0185574,0.0445817,0.055911,0.0718355,0.0360835,0.0188254,0.0127009,0.0699756},
{0.0890683,0.00538748,0.030639,0.195955,0.00979638,0.0177161,0.0122413,0.0439189,0.0433,0.0519788,0.0129031,0.018498,0.190437,0.0328772,0.0263208,0.0370614,0.042854,0.127892,0.00368064,0.0074749},
{0.024846,0.00458951,0.074297,0.0534399,0.00673798,0.231783,0.0267467,0.108869,0.0695202,0.0243092,0.00604431,0.0862321,0.0108379,0.0495179,0.0510762,0.0318472,0.0165701,0.111329,0.00304907,0.00835706},
{0.0244174,0.00489396,0.160963,0.0320525,0.0880157,0.0338822,0.0119483,0.00997069,0.0228931,0.0350139,0.00600504,0.0614953,0.0345717,0.0169126,0.0244614,0.10861,0.137054,0.0115409,0.135058,0.0402402},
{0.154992,0.00496452,0.0861302,0.111382,0.0692263,0.0399967,0.0126776,0.0221806,0.0389107,0.0312673,0.0100711,0.0250059,0.13838,0.0298271,0.0337106,0.0523295,0.0369363,0.028865,0.0107126,0.062435},
{0.0267537,0.00795484,0.197046,0.0111022,0.0196288,0.0146232,0.00741475,0.0459751,0.00626475,0.267154,0.016104,0.141628,0.0070205,0.00677252,0.00746031,0.0198844,0.0118344,0.172462,0.00651314,0.00640384},
{0.0395933,0.0219256,0.0130348,0.0309249,0.0857011,0.019066,0.135678,0.0213391,0.0153218,0.16065,0.0581322,0.211846,0.0300097,0.0230703,0.0218389,0.0291774,0.0174123,0.0221277,0.0174241,0.0257275},
{0.203487,0.00447037,0.0673934,0.101618,0.0540278,0.0347095,0.0174024,0.0284674,0.130739,0.042319,0.0114508,0.0425069,0.0131285,0.0526563,0.0399156,0.052014,0.0402123,0.0265381,0.00717916,0.029765},
{0.0460203,0.00739177,0.118815,0.0985432,0.0357327,0.0139629,0.0129906,0.0458196,0.0226783,0.0513222,0.147578,0.0188573,0.00786768,0.0736513,0.0181371,0.0265932,0.0331006,0.199814,0.00694103,0.0141836},
{0.125387,0.0182923,0.0202843,0.0297506,0.0330153,0.0272809,0.0375953,0.0489719,0.113517,0.117523,0.0252946,0.0232859,0.0130399,0.034582,0.0543283,0.0493626,0.0429494,0.0995598,0.0156714,0.0703077},
{0.0940518,0.0138662,0.0195902,0.0204566,0.0515227,0.0850886,0.0304329,0.0640517,0.0204133,0.0781691,0.0235832,0.031355,0.177292,0.0228116,0.0254432,0.0576935,0.0387261,0.0710895,0.0241932,0.0501698},
{0.0490678,0.00927317,0.0425124,0.100235,0.0383256,0.0250193,0.0260489,0.0633305,0.0656124,0.0827497,0.0203373,0.0330786,0.0143256,0.0476995,0.0725601,0.0427884,0.129719,0.0849623,0.0140863,0.0382677},
{0.0542128,0.00492328,0.0762294,0.102414,0.00642646,0.0410506,0.0175865,0.0137387,0.127494,0.018898,0.00715238,0.0363635,0.284654,0.0372132,0.0409982,0.0527735,0.048276,0.0188452,0.00402652,0.00672442},
{0.0398374,0.00501685,0.106609,0.0474481,0.0825963,0.219594,0.0130281,0.00676345,0.0325544,0.0144276,0.00572445,0.0377354,0.0121123,0.0288948,0.0435175,0.0808034,0.116884,0.012436,0.0113275,0.0826899},
{0.0539247,0.021219,0.0126118,0.0222247,0.034225,0.029726,0.0477668,0.0487313,0.0305514,0.147095,0.0192766,0.0243552,0.0227287,0.0316878,0.19374,0.0397249,0.0585798,0.0854704,0.018391,0.0579699},
{0.0656153,0.0125688,0.0189567,0.0368963,0.0264629,0.0481871,0.0167442,0.0369709,0.0254016,0.187732,0.0325312,0.038552,0.0155398,0.199939,0.0263241,0.0561504,0.0511057,0.0570488,0.00797307,0.0393001},
{0.012904,0.00170638,0.00735851,0.00403501,0.141313,0.0170429,0.0025491,0.0030635,0.0031484,0.013535,0.0021231,0.00888809,0.00782511,0.00311805,0.00646208,0.191536,0.193927,0.00394014,0.364466,0.0110595},
{0.0673302,0.00934423,0.05595,0.0644438,0.0295913,0.120099,0.0243047,0.0669625,0.0575751,0.10183,0.0258663,0.0471303,0.0183924,0.047575,0.0594031,0.0657573,0.0463845,0.0606094,0.00818042,0.0232702},
{0.265422,0.100319,0.0063246,0.00488977,0.0241983,0.0307479,0.00614797,0.0422519,0.00561867,0.226101,0.0212994,0.0111957,0.00528802,0.00625234,0.00983732,0.035293,0.0682114,0.0947064,0.0175499,0.0183447}};


#endif // BIOLOGICALSEQUENCES_H
